from textwrap import dedent

from kalamine import KeyboardLayout
from kalamine.template import osx_actions, osx_keymap, osx_terminators

from .util import get_layout_dict


def load_layout(filename: str) -> KeyboardLayout:
    return KeyboardLayout(get_layout_dict(filename))


def split(multiline_str: str):
    return dedent(multiline_str).lstrip().rstrip().splitlines()


EMPTY_KEYMAP = split(
    """
    <!-- Digits -->
    <key code="18"  output="&#x0010;" />
    <key code="19"  output="&#x0010;" />
    <key code="20"  output="&#x0010;" />
    <key code="21"  output="&#x0010;" />
    <key code="23"  output="&#x0010;" />
    <key code="22"  output="&#x0010;" />
    <key code="26"  output="&#x0010;" />
    <key code="28"  output="&#x0010;" />
    <key code="25"  output="&#x0010;" />
    <key code="29"  output="&#x0010;" />

    <!-- Letters, first row -->
    <key code="12"  output="&#x0010;" />
    <key code="13"  output="&#x0010;" />
    <key code="14"  output="&#x0010;" />
    <key code="15"  output="&#x0010;" />
    <key code="17"  output="&#x0010;" />
    <key code="16"  output="&#x0010;" />
    <key code="32"  output="&#x0010;" />
    <key code="34"  output="&#x0010;" />
    <key code="31"  output="&#x0010;" />
    <key code="35"  output="&#x0010;" />

    <!-- Letters, second row -->
    <key code="0"   output="&#x0010;" />
    <key code="1"   output="&#x0010;" />
    <key code="2"   output="&#x0010;" />
    <key code="3"   output="&#x0010;" />
    <key code="5"   output="&#x0010;" />
    <key code="4"   output="&#x0010;" />
    <key code="38"  output="&#x0010;" />
    <key code="40"  output="&#x0010;" />
    <key code="37"  output="&#x0010;" />
    <key code="41"  output="&#x0010;" />

    <!-- Letters, third row -->
    <key code="6"   output="&#x0010;" />
    <key code="7"   output="&#x0010;" />
    <key code="8"   output="&#x0010;" />
    <key code="9"   output="&#x0010;" />
    <key code="11"  output="&#x0010;" />
    <key code="45"  output="&#x0010;" />
    <key code="46"  output="&#x0010;" />
    <key code="43"  output="&#x0010;" />
    <key code="47"  output="&#x0010;" />
    <key code="44"  output="&#x0010;" />

    <!-- Pinky keys -->
    <key code="27"  output="&#x0010;" />
    <key code="24"  output="&#x0010;" />
    <key code="33"  output="&#x0010;" />
    <key code="30"  output="&#x0010;" />
    <key code="39"  output="&#x0010;" />
    <key code="50"  output="&#x0010;" />
    <key code="42"  output="&#x0010;" />
    <key code="10"  output="&#x0010;" />

    <!-- Space bar -->
    <key code="49"  output="&#x0010;" />
    """
)


def test_ansi():
    layout = load_layout("ansi")

    keymap = osx_keymap(layout)

    assert len(keymap[0]) == 60
    assert keymap[0] == split(
        """
        <!-- Digits -->
        <key code="18"  output="1" />
        <key code="19"  output="2" />
        <key code="20"  output="3" />
        <key code="21"  output="4" />
        <key code="23"  output="5" />
        <key code="22"  output="6" />
        <key code="26"  output="7" />
        <key code="28"  output="8" />
        <key code="25"  output="9" />
        <key code="29"  output="0" />

        <!-- Letters, first row -->
        <key code="12"  output="q" />
        <key code="13"  output="w" />
        <key code="14"  output="e" />
        <key code="15"  output="r" />
        <key code="17"  output="t" />
        <key code="16"  output="y" />
        <key code="32"  output="u" />
        <key code="34"  output="i" />
        <key code="31"  output="o" />
        <key code="35"  output="p" />

        <!-- Letters, second row -->
        <key code="0"   output="a" />
        <key code="1"   output="s" />
        <key code="2"   output="d" />
        <key code="3"   output="f" />
        <key code="5"   output="g" />
        <key code="4"   output="h" />
        <key code="38"  output="j" />
        <key code="40"  output="k" />
        <key code="37"  output="l" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   output="z" />
        <key code="7"   output="x" />
        <key code="8"   output="c" />
        <key code="9"   output="v" />
        <key code="11"  output="b" />
        <key code="45"  output="n" />
        <key code="46"  output="m" />
        <key code="43"  output="," />
        <key code="47"  output="." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  output="-" />
        <key code="24"  output="=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  output="'" />
        <key code="50"  output="`" />
        <key code="42"  output="\\" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[1]) == 60
    assert keymap[1] == split(
        """
        <!-- Digits -->
        <key code="18"  output="!" />
        <key code="19"  output="@" />
        <key code="20"  output="#" />
        <key code="21"  output="$" />
        <key code="23"  output="%" />
        <key code="22"  output="^" />
        <key code="26"  output="&#x0026;" />
        <key code="28"  output="*" />
        <key code="25"  output="(" />
        <key code="29"  output=")" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  output="W" />
        <key code="14"  output="E" />
        <key code="15"  output="R" />
        <key code="17"  output="T" />
        <key code="16"  output="Y" />
        <key code="32"  output="U" />
        <key code="34"  output="I" />
        <key code="31"  output="O" />
        <key code="35"  output="P" />

        <!-- Letters, second row -->
        <key code="0"   output="A" />
        <key code="1"   output="S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   output="G" />
        <key code="4"   output="H" />
        <key code="38"  output="J" />
        <key code="40"  output="K" />
        <key code="37"  output="L" />
        <key code="41"  output=":" />

        <!-- Letters, third row -->
        <key code="6"   output="Z" />
        <key code="7"   output="X" />
        <key code="8"   output="C" />
        <key code="9"   output="V" />
        <key code="11"  output="B" />
        <key code="45"  output="N" />
        <key code="46"  output="M" />
        <key code="43"  output="&#x003c;" />
        <key code="47"  output="&#x003e;" />
        <key code="44"  output="?" />

        <!-- Pinky keys -->
        <key code="27"  output="_" />
        <key code="24"  output="+" />
        <key code="33"  output="{" />
        <key code="30"  output="}" />
        <key code="39"  output="&#x0022;" />
        <key code="50"  output="~" />
        <key code="42"  output="|" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[2]) == 60
    assert keymap[2] == split(
        """
        <!-- Digits -->
        <key code="18"  output="1" />
        <key code="19"  output="2" />
        <key code="20"  output="3" />
        <key code="21"  output="4" />
        <key code="23"  output="5" />
        <key code="22"  output="6" />
        <key code="26"  output="7" />
        <key code="28"  output="8" />
        <key code="25"  output="9" />
        <key code="29"  output="0" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  output="W" />
        <key code="14"  output="E" />
        <key code="15"  output="R" />
        <key code="17"  output="T" />
        <key code="16"  output="Y" />
        <key code="32"  output="U" />
        <key code="34"  output="I" />
        <key code="31"  output="O" />
        <key code="35"  output="P" />

        <!-- Letters, second row -->
        <key code="0"   output="A" />
        <key code="1"   output="S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   output="G" />
        <key code="4"   output="H" />
        <key code="38"  output="J" />
        <key code="40"  output="K" />
        <key code="37"  output="L" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   output="Z" />
        <key code="7"   output="X" />
        <key code="8"   output="C" />
        <key code="9"   output="V" />
        <key code="11"  output="B" />
        <key code="45"  output="N" />
        <key code="46"  output="M" />
        <key code="43"  output="," />
        <key code="47"  output="." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  output="-" />
        <key code="24"  output="=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  output="'" />
        <key code="50"  output="`" />
        <key code="42"  output="\\" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[3]) == 60
    assert keymap[3] == EMPTY_KEYMAP

    assert len(keymap[4]) == 60
    assert keymap[4] == EMPTY_KEYMAP

    actions = osx_actions(layout)
    assert actions[1:] == split(
        """
        <!-- Digits -->

        <!-- Letters, first row -->

        <!-- Letters, second row -->

        <!-- Letters, third row -->

        <!-- Pinky keys -->

        <!-- Space bar -->
        <action id="spce_x0020">
          <when state="none"       output="&#x0020;" />
        </action>
        <action id="spce_x00a0">
          <when state="none"       output="&#x00a0;" />
        </action>
        <action id="spce_x202f">
          <when state="none"       output="&#x202f;" />
        </action>
        """
    )

    terminators = osx_terminators(layout)
    assert len(terminators) == 0


def test_intl():
    layout = load_layout("intl")

    keymap = osx_keymap(layout)

    assert len(keymap[0]) == 60
    assert keymap[0] == split(
        """
        <!-- Digits -->
        <key code="18"  action="ae01_1" />
        <key code="19"  action="ae02_2" />
        <key code="20"  action="ae03_3" />
        <key code="21"  action="ae04_4" />
        <key code="23"  action="ae05_5" />
        <key code="22"  action="ae06_6" />
        <key code="26"  action="ae07_7" />
        <key code="28"  action="ae08_8" />
        <key code="25"  action="ae09_9" />
        <key code="29"  action="ae10_0" />

        <!-- Letters, first row -->
        <key code="12"  output="q" />
        <key code="13"  action="ad02_w" />
        <key code="14"  action="ad03_e" />
        <key code="15"  output="r" />
        <key code="17"  output="t" />
        <key code="16"  action="ad06_y" />
        <key code="32"  action="ad07_u" />
        <key code="34"  action="ad08_i" />
        <key code="31"  action="ad09_o" />
        <key code="35"  output="p" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_a" />
        <key code="1"   action="ac02_s" />
        <key code="2"   output="d" />
        <key code="3"   output="f" />
        <key code="5"   action="ac05_g" />
        <key code="4"   action="ac06_h" />
        <key code="38"  action="ac07_j" />
        <key code="40"  output="k" />
        <key code="37"  output="l" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_z" />
        <key code="7"   action="ab02_x" />
        <key code="8"   action="ab03_c" />
        <key code="9"   action="ab04_v" />
        <key code="11"  output="b" />
        <key code="45"  action="ab06_n" />
        <key code="46"  output="m" />
        <key code="43"  output="," />
        <key code="47"  action="ab09_." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  action="ae11_-" />
        <key code="24"  action="ae12_=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  action="dead_1dk" />
        <key code="50"  action="dead_grave" />
        <key code="42"  output="\\" />
        <key code="10"  output="\\" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[1]) == 60
    assert keymap[1] == split(
        """
        <!-- Digits -->
        <key code="18"  output="!" />
        <key code="19"  output="@" />
        <key code="20"  output="#" />
        <key code="21"  output="$" />
        <key code="23"  output="%" />
        <key code="22"  action="dead_circumflex" />
        <key code="26"  output="&#x0026;" />
        <key code="28"  output="*" />
        <key code="25"  action="ae09_(" />
        <key code="29"  action="ae10_)" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  action="ad02_W" />
        <key code="14"  action="ad03_E" />
        <key code="15"  output="R" />
        <key code="17"  output="T" />
        <key code="16"  action="ad06_Y" />
        <key code="32"  action="ad07_U" />
        <key code="34"  action="ad08_I" />
        <key code="31"  action="ad09_O" />
        <key code="35"  output="P" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_A" />
        <key code="1"   action="ac02_S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   action="ac05_G" />
        <key code="4"   action="ac06_H" />
        <key code="38"  action="ac07_J" />
        <key code="40"  output="K" />
        <key code="37"  output="L" />
        <key code="41"  output=":" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_Z" />
        <key code="7"   action="ab02_X" />
        <key code="8"   action="ab03_C" />
        <key code="9"   action="ab04_V" />
        <key code="11"  output="B" />
        <key code="45"  action="ab06_N" />
        <key code="46"  output="M" />
        <key code="43"  action="ab08_x003c" />
        <key code="47"  action="ab09_x003e" />
        <key code="44"  output="?" />

        <!-- Pinky keys -->
        <key code="27"  output="_" />
        <key code="24"  action="ae12_+" />
        <key code="33"  output="{" />
        <key code="30"  output="}" />
        <key code="39"  action="dead_diaeresis" />
        <key code="50"  action="dead_tilde" />
        <key code="42"  output="|" />
        <key code="10"  output="|" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[2]) == 60
    assert keymap[2] == split(
        """
        <!-- Digits -->
        <key code="18"  action="ae01_1" />
        <key code="19"  action="ae02_2" />
        <key code="20"  action="ae03_3" />
        <key code="21"  action="ae04_4" />
        <key code="23"  action="ae05_5" />
        <key code="22"  action="ae06_6" />
        <key code="26"  action="ae07_7" />
        <key code="28"  action="ae08_8" />
        <key code="25"  action="ae09_9" />
        <key code="29"  action="ae10_0" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  action="ad02_W" />
        <key code="14"  action="ad03_E" />
        <key code="15"  output="R" />
        <key code="17"  output="T" />
        <key code="16"  action="ad06_Y" />
        <key code="32"  action="ad07_U" />
        <key code="34"  action="ad08_I" />
        <key code="31"  action="ad09_O" />
        <key code="35"  output="P" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_A" />
        <key code="1"   action="ac02_S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   action="ac05_G" />
        <key code="4"   action="ac06_H" />
        <key code="38"  action="ac07_J" />
        <key code="40"  output="K" />
        <key code="37"  output="L" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_Z" />
        <key code="7"   action="ab02_X" />
        <key code="8"   action="ab03_C" />
        <key code="9"   action="ab04_V" />
        <key code="11"  output="B" />
        <key code="45"  action="ab06_N" />
        <key code="46"  output="M" />
        <key code="43"  output="," />
        <key code="47"  action="ab09_." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  action="ae11_-" />
        <key code="24"  action="ae12_=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  action="dead_1dk" />
        <key code="50"  action="dead_grave" />
        <key code="42"  output="\\" />
        <key code="10"  output="\\" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[3]) == 60
    assert keymap[3] == EMPTY_KEYMAP

    assert len(keymap[4]) == 60
    assert keymap[4] == EMPTY_KEYMAP

    actions = osx_actions(layout)
    assert actions == split(
        """
        <action id="dead_1dk">
          <when state="none" next="1dk" />
        </action>
        <action id="dead_grave">
          <when state="none" next="grave" />
        </action>
        <action id="dead_circumflex">
          <when state="none" next="circumflex" />
        </action>
        <action id="dead_tilde">
          <when state="none" next="tilde" />
        </action>
        <action id="dead_diaeresis">
          <when state="none" next="diaeresis" />
        </action>

        <!-- Digits -->
        <action id="ae01_1">
          <when state="none"       output="1" />
          <when state="circumflex" output="¹" />
        </action>
        <action id="ae02_2">
          <when state="none"       output="2" />
          <when state="circumflex" output="²" />
        </action>
        <action id="ae03_3">
          <when state="none"       output="3" />
          <when state="circumflex" output="³" />
        </action>
        <action id="ae04_4">
          <when state="none"       output="4" />
          <when state="circumflex" output="⁴" />
        </action>
        <action id="ae05_5">
          <when state="none"       output="5" />
          <when state="circumflex" output="⁵" />
        </action>
        <action id="ae06_6">
          <when state="none"       output="6" />
          <when state="circumflex" output="⁶" />
        </action>
        <action id="ae07_7">
          <when state="none"       output="7" />
          <when state="circumflex" output="⁷" />
        </action>
        <action id="ae08_8">
          <when state="none"       output="8" />
          <when state="circumflex" output="⁸" />
        </action>
        <action id="ae09_9">
          <when state="none"       output="9" />
          <when state="circumflex" output="⁹" />
        </action>
        <action id="ae09_(">
          <when state="none"       output="(" />
          <when state="circumflex" output="⁽" />
        </action>
        <action id="ae10_0">
          <when state="none"       output="0" />
          <when state="circumflex" output="⁰" />
        </action>
        <action id="ae10_)">
          <when state="none"       output=")" />
          <when state="circumflex" output="⁾" />
        </action>

        <!-- Letters, first row -->
        <action id="ad02_w">
          <when state="none"       output="w" />
          <when state="grave"      output="ẁ" />
          <when state="circumflex" output="ŵ" />
          <when state="diaeresis"  output="ẅ" />
        </action>
        <action id="ad02_W">
          <when state="none"       output="W" />
          <when state="grave"      output="Ẁ" />
          <when state="circumflex" output="Ŵ" />
          <when state="diaeresis"  output="Ẅ" />
        </action>
        <action id="ad03_e">
          <when state="none"       output="e" />
          <when state="1dk"        output="é" />
          <when state="grave"      output="è" />
          <when state="circumflex" output="ê" />
          <when state="tilde"      output="ẽ" />
          <when state="diaeresis"  output="ë" />
        </action>
        <action id="ad03_E">
          <when state="none"       output="E" />
          <when state="1dk"        output="É" />
          <when state="grave"      output="È" />
          <when state="circumflex" output="Ê" />
          <when state="tilde"      output="Ẽ" />
          <when state="diaeresis"  output="Ë" />
        </action>
        <action id="ad05_t">
          <when state="none"       output="t" />
          <when state="diaeresis"  output="ẗ" />
        </action>
        <action id="ad06_y">
          <when state="none"       output="y" />
          <when state="grave"      output="ỳ" />
          <when state="circumflex" output="ŷ" />
          <when state="tilde"      output="ỹ" />
          <when state="diaeresis"  output="ÿ" />
        </action>
        <action id="ad06_Y">
          <when state="none"       output="Y" />
          <when state="grave"      output="Ỳ" />
          <when state="circumflex" output="Ŷ" />
          <when state="tilde"      output="Ỹ" />
          <when state="diaeresis"  output="Ÿ" />
        </action>
        <action id="ad07_u">
          <when state="none"       output="u" />
          <when state="1dk"        output="ú" />
          <when state="grave"      output="ù" />
          <when state="circumflex" output="û" />
          <when state="tilde"      output="ũ" />
          <when state="diaeresis"  output="ü" />
        </action>
        <action id="ad07_U">
          <when state="none"       output="U" />
          <when state="1dk"        output="Ú" />
          <when state="grave"      output="Ù" />
          <when state="circumflex" output="Û" />
          <when state="tilde"      output="Ũ" />
          <when state="diaeresis"  output="Ü" />
        </action>
        <action id="ad08_i">
          <when state="none"       output="i" />
          <when state="1dk"        output="í" />
          <when state="grave"      output="ì" />
          <when state="circumflex" output="î" />
          <when state="tilde"      output="ĩ" />
          <when state="diaeresis"  output="ï" />
        </action>
        <action id="ad08_I">
          <when state="none"       output="I" />
          <when state="1dk"        output="Í" />
          <when state="grave"      output="Ì" />
          <when state="circumflex" output="Î" />
          <when state="tilde"      output="Ĩ" />
          <when state="diaeresis"  output="Ï" />
        </action>
        <action id="ad09_o">
          <when state="none"       output="o" />
          <when state="1dk"        output="ó" />
          <when state="grave"      output="ò" />
          <when state="circumflex" output="ô" />
          <when state="tilde"      output="õ" />
          <when state="diaeresis"  output="ö" />
        </action>
        <action id="ad09_O">
          <when state="none"       output="O" />
          <when state="1dk"        output="Ó" />
          <when state="grave"      output="Ò" />
          <when state="circumflex" output="Ô" />
          <when state="tilde"      output="Õ" />
          <when state="diaeresis"  output="Ö" />
        </action>

        <!-- Letters, second row -->
        <action id="ac01_a">
          <when state="none"       output="a" />
          <when state="1dk"        output="á" />
          <when state="grave"      output="à" />
          <when state="circumflex" output="â" />
          <when state="tilde"      output="ã" />
          <when state="diaeresis"  output="ä" />
        </action>
        <action id="ac01_A">
          <when state="none"       output="A" />
          <when state="1dk"        output="Á" />
          <when state="grave"      output="À" />
          <when state="circumflex" output="Â" />
          <when state="tilde"      output="Ã" />
          <when state="diaeresis"  output="Ä" />
        </action>
        <action id="ac02_s">
          <when state="none"       output="s" />
          <when state="circumflex" output="ŝ" />
        </action>
        <action id="ac02_S">
          <when state="none"       output="S" />
          <when state="circumflex" output="Ŝ" />
        </action>
        <action id="ac05_g">
          <when state="none"       output="g" />
          <when state="circumflex" output="ĝ" />
        </action>
        <action id="ac05_G">
          <when state="none"       output="G" />
          <when state="circumflex" output="Ĝ" />
        </action>
        <action id="ac06_h">
          <when state="none"       output="h" />
          <when state="circumflex" output="ĥ" />
          <when state="diaeresis"  output="ḧ" />
        </action>
        <action id="ac06_H">
          <when state="none"       output="H" />
          <when state="circumflex" output="Ĥ" />
          <when state="diaeresis"  output="Ḧ" />
        </action>
        <action id="ac07_j">
          <when state="none"       output="j" />
          <when state="circumflex" output="ĵ" />
        </action>
        <action id="ac07_J">
          <when state="none"       output="J" />
          <when state="circumflex" output="Ĵ" />
        </action>

        <!-- Letters, third row -->
        <action id="ab01_z">
          <when state="none"       output="z" />
          <when state="circumflex" output="ẑ" />
        </action>
        <action id="ab01_Z">
          <when state="none"       output="Z" />
          <when state="circumflex" output="Ẑ" />
        </action>
        <action id="ab02_x">
          <when state="none"       output="x" />
          <when state="diaeresis"  output="ẍ" />
        </action>
        <action id="ab02_X">
          <when state="none"       output="X" />
          <when state="diaeresis"  output="Ẍ" />
        </action>
        <action id="ab03_c">
          <when state="none"       output="c" />
          <when state="1dk"        output="ç" />
          <when state="circumflex" output="ĉ" />
        </action>
        <action id="ab03_C">
          <when state="none"       output="C" />
          <when state="1dk"        output="Ç" />
          <when state="circumflex" output="Ĉ" />
        </action>
        <action id="ab04_v">
          <when state="none"       output="v" />
          <when state="tilde"      output="ṽ" />
        </action>
        <action id="ab04_V">
          <when state="none"       output="V" />
          <when state="tilde"      output="Ṽ" />
        </action>
        <action id="ab06_n">
          <when state="none"       output="n" />
          <when state="grave"      output="ǹ" />
          <when state="tilde"      output="ñ" />
        </action>
        <action id="ab06_N">
          <when state="none"       output="N" />
          <when state="grave"      output="Ǹ" />
          <when state="tilde"      output="Ñ" />
        </action>
        <action id="ab08_x003c">
          <when state="none"       output="&#x003c;" />
          <when state="tilde"      output="≲" />
        </action>
        <action id="ab09_.">
          <when state="none"       output="." />
          <when state="1dk"        output="…" />
        </action>
        <action id="ab09_x003e">
          <when state="none"       output="&#x003e;" />
          <when state="tilde"      output="≳" />
        </action>

        <!-- Pinky keys -->
        <action id="ae11_-">
          <when state="none"       output="-" />
          <when state="circumflex" output="⁻" />
        </action>
        <action id="ae12_=">
          <when state="none"       output="=" />
          <when state="circumflex" output="⁼" />
          <when state="tilde"      output="≃" />
        </action>
        <action id="ae12_+">
          <when state="none"       output="+" />
          <when state="circumflex" output="⁺" />
        </action>

        <!-- Space bar -->
        <action id="spce_x0020">
          <when state="none"       output="&#x0020;" />
          <when state="1dk"        output="'" />
          <when state="grave"      output="`" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        <action id="spce_x00a0">
          <when state="none"       output="&#x00a0;" />
          <when state="1dk"        output="'" />
          <when state="grave"      output="`" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        <action id="spce_x202f">
          <when state="none"       output="&#x202f;" />
          <when state="1dk"        output="'" />
          <when state="grave"      output="`" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        """
    )

    terminators = osx_terminators(layout)
    assert len(terminators) == 5
    assert terminators == split(
        """
        <when state="1dk"        output="\'" />
        <when state="grave"      output="`" />
        <when state="circumflex" output="^" />
        <when state="tilde"      output="~" />
        <when state="diaeresis"  output="¨" />
        """
    )


def test_prog():
    layout = load_layout("prog")

    keymap = osx_keymap(layout)

    assert len(keymap[0]) == 60
    assert keymap[0] == split(
        """
        <!-- Digits -->
        <key code="18"  action="ae01_1" />
        <key code="19"  action="ae02_2" />
        <key code="20"  action="ae03_3" />
        <key code="21"  action="ae04_4" />
        <key code="23"  action="ae05_5" />
        <key code="22"  action="ae06_6" />
        <key code="26"  action="ae07_7" />
        <key code="28"  action="ae08_8" />
        <key code="25"  action="ae09_9" />
        <key code="29"  action="ae10_0" />

        <!-- Letters, first row -->
        <key code="12"  output="q" />
        <key code="13"  action="ad02_w" />
        <key code="14"  action="ad03_e" />
        <key code="15"  action="ad04_r" />
        <key code="17"  output="t" />
        <key code="16"  action="ad06_y" />
        <key code="32"  action="ad07_u" />
        <key code="34"  action="ad08_i" />
        <key code="31"  action="ad09_o" />
        <key code="35"  action="ad10_p" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_a" />
        <key code="1"   action="ac02_s" />
        <key code="2"   output="d" />
        <key code="3"   output="f" />
        <key code="5"   action="ac05_g" />
        <key code="4"   action="ac06_h" />
        <key code="38"  action="ac07_j" />
        <key code="40"  action="ac08_k" />
        <key code="37"  action="ac09_l" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_z" />
        <key code="7"   action="ab02_x" />
        <key code="8"   action="ab03_c" />
        <key code="9"   action="ab04_v" />
        <key code="11"  output="b" />
        <key code="45"  action="ab06_n" />
        <key code="46"  action="ab07_m" />
        <key code="43"  output="," />
        <key code="47"  output="." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  action="ae11_-" />
        <key code="24"  action="ae12_=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  output="'" />
        <key code="50"  output="`" />
        <key code="42"  output="\\" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[1]) == 60
    assert keymap[1] == split(
        """
        <!-- Digits -->
        <key code="18"  output="!" />
        <key code="19"  output="@" />
        <key code="20"  output="#" />
        <key code="21"  output="$" />
        <key code="23"  output="%" />
        <key code="22"  output="^" />
        <key code="26"  output="&#x0026;" />
        <key code="28"  output="*" />
        <key code="25"  action="ae09_(" />
        <key code="29"  action="ae10_)" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  action="ad02_W" />
        <key code="14"  action="ad03_E" />
        <key code="15"  action="ad04_R" />
        <key code="17"  output="T" />
        <key code="16"  action="ad06_Y" />
        <key code="32"  action="ad07_U" />
        <key code="34"  action="ad08_I" />
        <key code="31"  action="ad09_O" />
        <key code="35"  action="ad10_P" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_A" />
        <key code="1"   action="ac02_S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   action="ac05_G" />
        <key code="4"   action="ac06_H" />
        <key code="38"  action="ac07_J" />
        <key code="40"  action="ac08_K" />
        <key code="37"  action="ac09_L" />
        <key code="41"  output=":" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_Z" />
        <key code="7"   action="ab02_X" />
        <key code="8"   action="ab03_C" />
        <key code="9"   action="ab04_V" />
        <key code="11"  output="B" />
        <key code="45"  action="ab06_N" />
        <key code="46"  action="ab07_M" />
        <key code="43"  action="ab08_x003c" />
        <key code="47"  action="ab09_x003e" />
        <key code="44"  output="?" />

        <!-- Pinky keys -->
        <key code="27"  output="_" />
        <key code="24"  action="ae12_+" />
        <key code="33"  output="{" />
        <key code="30"  output="}" />
        <key code="39"  output="&#x0022;" />
        <key code="50"  output="~" />
        <key code="42"  output="|" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[2]) == 60
    assert keymap[2] == split(
        """
        <!-- Digits -->
        <key code="18"  action="ae01_1" />
        <key code="19"  action="ae02_2" />
        <key code="20"  action="ae03_3" />
        <key code="21"  action="ae04_4" />
        <key code="23"  action="ae05_5" />
        <key code="22"  action="ae06_6" />
        <key code="26"  action="ae07_7" />
        <key code="28"  action="ae08_8" />
        <key code="25"  action="ae09_9" />
        <key code="29"  action="ae10_0" />

        <!-- Letters, first row -->
        <key code="12"  output="Q" />
        <key code="13"  action="ad02_W" />
        <key code="14"  action="ad03_E" />
        <key code="15"  action="ad04_R" />
        <key code="17"  output="T" />
        <key code="16"  action="ad06_Y" />
        <key code="32"  action="ad07_U" />
        <key code="34"  action="ad08_I" />
        <key code="31"  action="ad09_O" />
        <key code="35"  action="ad10_P" />

        <!-- Letters, second row -->
        <key code="0"   action="ac01_A" />
        <key code="1"   action="ac02_S" />
        <key code="2"   output="D" />
        <key code="3"   output="F" />
        <key code="5"   action="ac05_G" />
        <key code="4"   action="ac06_H" />
        <key code="38"  action="ac07_J" />
        <key code="40"  action="ac08_K" />
        <key code="37"  action="ac09_L" />
        <key code="41"  output=";" />

        <!-- Letters, third row -->
        <key code="6"   action="ab01_Z" />
        <key code="7"   action="ab02_X" />
        <key code="8"   action="ab03_C" />
        <key code="9"   action="ab04_V" />
        <key code="11"  output="B" />
        <key code="45"  action="ab06_N" />
        <key code="46"  action="ab07_M" />
        <key code="43"  output="," />
        <key code="47"  output="." />
        <key code="44"  output="/" />

        <!-- Pinky keys -->
        <key code="27"  action="ae11_-" />
        <key code="24"  action="ae12_=" />
        <key code="33"  output="[" />
        <key code="30"  output="]" />
        <key code="39"  output="'" />
        <key code="50"  output="`" />
        <key code="42"  output="\\" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[3]) == 60
    assert keymap[3] == split(
        """
        <!-- Digits -->
        <key code="18"  output="!" />
        <key code="19"  action="ae02_(" />
        <key code="20"  action="ae03_)" />
        <key code="21"  output="'" />
        <key code="23"  output="&#x0022;" />
        <key code="22"  action="dead_circumflex" />
        <key code="26"  action="ae07_7" />
        <key code="28"  action="ae08_8" />
        <key code="25"  action="ae09_9" />
        <key code="29"  output="/" />

        <!-- Letters, first row -->
        <key code="12"  action="ad01_=" />
        <key code="13"  action="ad02_x003c" />
        <key code="14"  action="ad03_x003e" />
        <key code="15"  action="ad04_-" />
        <key code="17"  action="ad05_+" />
        <key code="16"  output="&#x0010;" />
        <key code="32"  action="ad07_4" />
        <key code="34"  action="ad08_5" />
        <key code="31"  action="ad09_6" />
        <key code="35"  output="*" />

        <!-- Letters, second row -->
        <key code="0"   output="{" />
        <key code="1"   output="[" />
        <key code="2"   output="]" />
        <key code="3"   output="}" />
        <key code="5"   output="/" />
        <key code="4"   output="&#x0010;" />
        <key code="38"  action="ac07_1" />
        <key code="40"  action="ac08_2" />
        <key code="37"  action="ac09_3" />
        <key code="41"  action="ac10_-" />

        <!-- Letters, third row -->
        <key code="6"   output="~" />
        <key code="7"   output="`" />
        <key code="8"   output="|" />
        <key code="9"   output="_" />
        <key code="11"  output="\\" />
        <key code="45"  output="&#x0010;" />
        <key code="46"  action="ab07_0" />
        <key code="43"  output="," />
        <key code="47"  output="." />
        <key code="44"  action="ab10_+" />

        <!-- Pinky keys -->
        <key code="27"  output="&#x0010;" />
        <key code="24"  output="&#x0010;" />
        <key code="33"  output="&#x0010;" />
        <key code="30"  output="&#x0010;" />
        <key code="39"  action="dead_acute" />
        <key code="50"  action="dead_grave" />
        <key code="42"  output="&#x0010;" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    assert len(keymap[4]) == 60
    assert keymap[4] == split(
        """
        <!-- Digits -->
        <key code="18"  output="&#x0010;" />
        <key code="19"  output="&#x0010;" />
        <key code="20"  output="&#x0010;" />
        <key code="21"  output="&#x0010;" />
        <key code="23"  output="&#x0010;" />
        <key code="22"  output="&#x0010;" />
        <key code="26"  output="&#x0010;" />
        <key code="28"  output="&#x0010;" />
        <key code="25"  output="&#x0010;" />
        <key code="29"  output="&#x0010;" />

        <!-- Letters, first row -->
        <key code="12"  output="&#x0010;" />
        <key code="13"  output="≤" />
        <key code="14"  output="≥" />
        <key code="15"  output="&#x0010;" />
        <key code="17"  output="&#x0010;" />
        <key code="16"  output="&#x0010;" />
        <key code="32"  output="&#x0010;" />
        <key code="34"  output="&#x0010;" />
        <key code="31"  output="&#x0010;" />
        <key code="35"  output="&#x0010;" />

        <!-- Letters, second row -->
        <key code="0"   output="&#x0010;" />
        <key code="1"   output="&#x0010;" />
        <key code="2"   output="&#x0010;" />
        <key code="3"   output="&#x0010;" />
        <key code="5"   output="&#x0010;" />
        <key code="4"   output="&#x0010;" />
        <key code="38"  output="&#x0010;" />
        <key code="40"  output="&#x0010;" />
        <key code="37"  output="&#x0010;" />
        <key code="41"  output="&#x0010;" />

        <!-- Letters, third row -->
        <key code="6"   output="&#x0010;" />
        <key code="7"   output="&#x0010;" />
        <key code="8"   output="¦" />
        <key code="9"   output="&#x0010;" />
        <key code="11"  output="&#x0010;" />
        <key code="45"  output="&#x0010;" />
        <key code="46"  output="&#x0010;" />
        <key code="43"  output="&#x0010;" />
        <key code="47"  output="&#x0010;" />
        <key code="44"  output="&#x0010;" />

        <!-- Pinky keys -->
        <key code="27"  output="&#x0010;" />
        <key code="24"  output="&#x0010;" />
        <key code="33"  output="&#x0010;" />
        <key code="30"  output="&#x0010;" />
        <key code="39"  action="dead_diaeresis" />
        <key code="50"  action="dead_tilde" />
        <key code="42"  output="&#x0010;" />
        <key code="10"  output="&#x0010;" />

        <!-- Space bar -->
        <key code="49"  action="spce_x0020" />
        """
    )

    actions = osx_actions(layout)
    assert actions == split(
        """
        <action id="dead_grave">
          <when state="none" next="grave" />
        </action>
        <action id="dead_acute">
          <when state="none" next="acute" />
        </action>
        <action id="dead_circumflex">
          <when state="none" next="circumflex" />
        </action>
        <action id="dead_tilde">
          <when state="none" next="tilde" />
        </action>
        <action id="dead_diaeresis">
          <when state="none" next="diaeresis" />
        </action>

        <!-- Digits -->
        <action id="ae01_1">
          <when state="none"       output="1" />
          <when state="circumflex" output="¹" />
        </action>
        <action id="ae02_2">
          <when state="none"       output="2" />
          <when state="circumflex" output="²" />
        </action>
        <action id="ae02_(">
          <when state="none"       output="(" />
          <when state="circumflex" output="⁽" />
        </action>
        <action id="ae03_3">
          <when state="none"       output="3" />
          <when state="circumflex" output="³" />
        </action>
        <action id="ae03_)">
          <when state="none"       output=")" />
          <when state="circumflex" output="⁾" />
        </action>
        <action id="ae04_4">
          <when state="none"       output="4" />
          <when state="circumflex" output="⁴" />
        </action>
        <action id="ae05_5">
          <when state="none"       output="5" />
          <when state="circumflex" output="⁵" />
        </action>
        <action id="ae06_6">
          <when state="none"       output="6" />
          <when state="circumflex" output="⁶" />
        </action>
        <action id="ae07_7">
          <when state="none"       output="7" />
          <when state="circumflex" output="⁷" />
        </action>
        <action id="ae08_8">
          <when state="none"       output="8" />
          <when state="circumflex" output="⁸" />
        </action>
        <action id="ae09_9">
          <when state="none"       output="9" />
          <when state="circumflex" output="⁹" />
        </action>
        <action id="ae09_(">
          <when state="none"       output="(" />
          <when state="circumflex" output="⁽" />
        </action>
        <action id="ae10_0">
          <when state="none"       output="0" />
          <when state="circumflex" output="⁰" />
        </action>
        <action id="ae10_)">
          <when state="none"       output=")" />
          <when state="circumflex" output="⁾" />
        </action>

        <!-- Letters, first row -->
        <action id="ad01_=">
          <when state="none"       output="=" />
          <when state="circumflex" output="⁼" />
          <when state="tilde"      output="≃" />
        </action>
        <action id="ad02_w">
          <when state="none"       output="w" />
          <when state="grave"      output="ẁ" />
          <when state="acute"      output="ẃ" />
          <when state="circumflex" output="ŵ" />
          <when state="diaeresis"  output="ẅ" />
        </action>
        <action id="ad02_W">
          <when state="none"       output="W" />
          <when state="grave"      output="Ẁ" />
          <when state="acute"      output="Ẃ" />
          <when state="circumflex" output="Ŵ" />
          <when state="diaeresis"  output="Ẅ" />
        </action>
        <action id="ad02_x003c">
          <when state="none"       output="&#x003c;" />
          <when state="tilde"      output="≲" />
        </action>
        <action id="ad03_e">
          <when state="none"       output="e" />
          <when state="grave"      output="è" />
          <when state="acute"      output="é" />
          <when state="circumflex" output="ê" />
          <when state="tilde"      output="ẽ" />
          <when state="diaeresis"  output="ë" />
        </action>
        <action id="ad03_E">
          <when state="none"       output="E" />
          <when state="grave"      output="È" />
          <when state="acute"      output="É" />
          <when state="circumflex" output="Ê" />
          <when state="tilde"      output="Ẽ" />
          <when state="diaeresis"  output="Ë" />
        </action>
        <action id="ad03_x003e">
          <when state="none"       output="&#x003e;" />
          <when state="tilde"      output="≳" />
        </action>
        <action id="ad04_r">
          <when state="none"       output="r" />
          <when state="acute"      output="ŕ" />
        </action>
        <action id="ad04_R">
          <when state="none"       output="R" />
          <when state="acute"      output="Ŕ" />
        </action>
        <action id="ad04_-">
          <when state="none"       output="-" />
          <when state="circumflex" output="⁻" />
        </action>
        <action id="ad05_t">
          <when state="none"       output="t" />
          <when state="diaeresis"  output="ẗ" />
        </action>
        <action id="ad05_+">
          <when state="none"       output="+" />
          <when state="circumflex" output="⁺" />
        </action>
        <action id="ad06_y">
          <when state="none"       output="y" />
          <when state="grave"      output="ỳ" />
          <when state="acute"      output="ý" />
          <when state="circumflex" output="ŷ" />
          <when state="tilde"      output="ỹ" />
          <when state="diaeresis"  output="ÿ" />
        </action>
        <action id="ad06_Y">
          <when state="none"       output="Y" />
          <when state="grave"      output="Ỳ" />
          <when state="acute"      output="Ý" />
          <when state="circumflex" output="Ŷ" />
          <when state="tilde"      output="Ỹ" />
          <when state="diaeresis"  output="Ÿ" />
        </action>
        <action id="ad07_u">
          <when state="none"       output="u" />
          <when state="grave"      output="ù" />
          <when state="acute"      output="ú" />
          <when state="circumflex" output="û" />
          <when state="tilde"      output="ũ" />
          <when state="diaeresis"  output="ü" />
        </action>
        <action id="ad07_U">
          <when state="none"       output="U" />
          <when state="grave"      output="Ù" />
          <when state="acute"      output="Ú" />
          <when state="circumflex" output="Û" />
          <when state="tilde"      output="Ũ" />
          <when state="diaeresis"  output="Ü" />
        </action>
        <action id="ad07_4">
          <when state="none"       output="4" />
          <when state="circumflex" output="⁴" />
        </action>
        <action id="ad08_i">
          <when state="none"       output="i" />
          <when state="grave"      output="ì" />
          <when state="acute"      output="í" />
          <when state="circumflex" output="î" />
          <when state="tilde"      output="ĩ" />
          <when state="diaeresis"  output="ï" />
        </action>
        <action id="ad08_I">
          <when state="none"       output="I" />
          <when state="grave"      output="Ì" />
          <when state="acute"      output="Í" />
          <when state="circumflex" output="Î" />
          <when state="tilde"      output="Ĩ" />
          <when state="diaeresis"  output="Ï" />
        </action>
        <action id="ad08_5">
          <when state="none"       output="5" />
          <when state="circumflex" output="⁵" />
        </action>
        <action id="ad09_o">
          <when state="none"       output="o" />
          <when state="grave"      output="ò" />
          <when state="acute"      output="ó" />
          <when state="circumflex" output="ô" />
          <when state="tilde"      output="õ" />
          <when state="diaeresis"  output="ö" />
        </action>
        <action id="ad09_O">
          <when state="none"       output="O" />
          <when state="grave"      output="Ò" />
          <when state="acute"      output="Ó" />
          <when state="circumflex" output="Ô" />
          <when state="tilde"      output="Õ" />
          <when state="diaeresis"  output="Ö" />
        </action>
        <action id="ad09_6">
          <when state="none"       output="6" />
          <when state="circumflex" output="⁶" />
        </action>
        <action id="ad10_p">
          <when state="none"       output="p" />
          <when state="acute"      output="ṕ" />
        </action>
        <action id="ad10_P">
          <when state="none"       output="P" />
          <when state="acute"      output="Ṕ" />
        </action>

        <!-- Letters, second row -->
        <action id="ac01_a">
          <when state="none"       output="a" />
          <when state="grave"      output="à" />
          <when state="acute"      output="á" />
          <when state="circumflex" output="â" />
          <when state="tilde"      output="ã" />
          <when state="diaeresis"  output="ä" />
        </action>
        <action id="ac01_A">
          <when state="none"       output="A" />
          <when state="grave"      output="À" />
          <when state="acute"      output="Á" />
          <when state="circumflex" output="Â" />
          <when state="tilde"      output="Ã" />
          <when state="diaeresis"  output="Ä" />
        </action>
        <action id="ac02_s">
          <when state="none"       output="s" />
          <when state="acute"      output="ś" />
          <when state="circumflex" output="ŝ" />
        </action>
        <action id="ac02_S">
          <when state="none"       output="S" />
          <when state="acute"      output="Ś" />
          <when state="circumflex" output="Ŝ" />
        </action>
        <action id="ac05_g">
          <when state="none"       output="g" />
          <when state="acute"      output="ǵ" />
          <when state="circumflex" output="ĝ" />
        </action>
        <action id="ac05_G">
          <when state="none"       output="G" />
          <when state="acute"      output="Ǵ" />
          <when state="circumflex" output="Ĝ" />
        </action>
        <action id="ac06_h">
          <when state="none"       output="h" />
          <when state="circumflex" output="ĥ" />
          <when state="diaeresis"  output="ḧ" />
        </action>
        <action id="ac06_H">
          <when state="none"       output="H" />
          <when state="circumflex" output="Ĥ" />
          <when state="diaeresis"  output="Ḧ" />
        </action>
        <action id="ac07_j">
          <when state="none"       output="j" />
          <when state="circumflex" output="ĵ" />
        </action>
        <action id="ac07_J">
          <when state="none"       output="J" />
          <when state="circumflex" output="Ĵ" />
        </action>
        <action id="ac07_1">
          <when state="none"       output="1" />
          <when state="circumflex" output="¹" />
        </action>
        <action id="ac08_k">
          <when state="none"       output="k" />
          <when state="acute"      output="ḱ" />
        </action>
        <action id="ac08_K">
          <when state="none"       output="K" />
          <when state="acute"      output="Ḱ" />
        </action>
        <action id="ac08_2">
          <when state="none"       output="2" />
          <when state="circumflex" output="²" />
        </action>
        <action id="ac09_l">
          <when state="none"       output="l" />
          <when state="acute"      output="ĺ" />
        </action>
        <action id="ac09_L">
          <when state="none"       output="L" />
          <when state="acute"      output="Ĺ" />
        </action>
        <action id="ac09_3">
          <when state="none"       output="3" />
          <when state="circumflex" output="³" />
        </action>
        <action id="ac10_-">
          <when state="none"       output="-" />
          <when state="circumflex" output="⁻" />
        </action>

        <!-- Letters, third row -->
        <action id="ab01_z">
          <when state="none"       output="z" />
          <when state="acute"      output="ź" />
          <when state="circumflex" output="ẑ" />
        </action>
        <action id="ab01_Z">
          <when state="none"       output="Z" />
          <when state="acute"      output="Ź" />
          <when state="circumflex" output="Ẑ" />
        </action>
        <action id="ab02_x">
          <when state="none"       output="x" />
          <when state="diaeresis"  output="ẍ" />
        </action>
        <action id="ab02_X">
          <when state="none"       output="X" />
          <when state="diaeresis"  output="Ẍ" />
        </action>
        <action id="ab03_c">
          <when state="none"       output="c" />
          <when state="acute"      output="ć" />
          <when state="circumflex" output="ĉ" />
        </action>
        <action id="ab03_C">
          <when state="none"       output="C" />
          <when state="acute"      output="Ć" />
          <when state="circumflex" output="Ĉ" />
        </action>
        <action id="ab04_v">
          <when state="none"       output="v" />
          <when state="tilde"      output="ṽ" />
        </action>
        <action id="ab04_V">
          <when state="none"       output="V" />
          <when state="tilde"      output="Ṽ" />
        </action>
        <action id="ab06_n">
          <when state="none"       output="n" />
          <when state="grave"      output="ǹ" />
          <when state="acute"      output="ń" />
          <when state="tilde"      output="ñ" />
        </action>
        <action id="ab06_N">
          <when state="none"       output="N" />
          <when state="grave"      output="Ǹ" />
          <when state="acute"      output="Ń" />
          <when state="tilde"      output="Ñ" />
        </action>
        <action id="ab07_m">
          <when state="none"       output="m" />
          <when state="acute"      output="ḿ" />
        </action>
        <action id="ab07_M">
          <when state="none"       output="M" />
          <when state="acute"      output="Ḿ" />
        </action>
        <action id="ab07_0">
          <when state="none"       output="0" />
          <when state="circumflex" output="⁰" />
        </action>
        <action id="ab08_x003c">
          <when state="none"       output="&#x003c;" />
          <when state="tilde"      output="≲" />
        </action>
        <action id="ab09_x003e">
          <when state="none"       output="&#x003e;" />
          <when state="tilde"      output="≳" />
        </action>
        <action id="ab10_+">
          <when state="none"       output="+" />
          <when state="circumflex" output="⁺" />
        </action>

        <!-- Pinky keys -->
        <action id="ae11_-">
          <when state="none"       output="-" />
          <when state="circumflex" output="⁻" />
        </action>
        <action id="ae12_=">
          <when state="none"       output="=" />
          <when state="circumflex" output="⁼" />
          <when state="tilde"      output="≃" />
        </action>
        <action id="ae12_+">
          <when state="none"       output="+" />
          <when state="circumflex" output="⁺" />
        </action>

        <!-- Space bar -->
        <action id="spce_x0020">
          <when state="none"       output="&#x0020;" />
          <when state="grave"      output="`" />
          <when state="acute"      output="'" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        <action id="spce_x00a0">
          <when state="none"       output="&#x00a0;" />
          <when state="grave"      output="`" />
          <when state="acute"      output="'" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        <action id="spce_x202f">
          <when state="none"       output="&#x202f;" />
          <when state="grave"      output="`" />
          <when state="acute"      output="'" />
          <when state="circumflex" output="^" />
          <when state="tilde"      output="~" />
          <when state="diaeresis"  output="&#x0022;" />
        </action>
        """
    )

    terminators = osx_terminators(layout)
    assert len(terminators) == 5
    assert terminators == split(
        """
        <when state="grave"      output="`" />
        <when state="acute"      output="´" />
        <when state="circumflex" output="^" />
        <when state="tilde"      output="~" />
        <when state="diaeresis"  output="¨" />
        """
    )



import pkgutil
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, List

import yaml


def lines_to_text(lines: List[str], indent: str = "") -> str:
    """
    From a list lines of string, produce a string concatenating the elements
    of lines indented by prepending indent and followed by a new line.
    Example: lines_to_text(["one", "two", "three"], "  ") returns
    '  one\n  two\n  three'
    """
    out = ""
    for line in lines:
        if len(line):
            out += indent + line
        out += "\n"
    return out[:-1]


def text_to_lines(text: str) -> List[str]:
    """Split given text into lines"""
    return text.split("\n")


def load_data(filename: str) -> Dict:
    descriptor = pkgutil.get_data(__package__, f"data/{filename}.yaml")
    if not descriptor:
        return {}
    return yaml.safe_load(descriptor.decode("utf-8"))


class Layer(IntEnum):
    """A layer designation."""

    BASE = 0
    SHIFT = 1
    ODK = 2
    ODK_SHIFT = 3
    ALTGR = 4
    ALTGR_SHIFT = 5

    def next(self) -> "Layer":
        """The next layer in the layer ordering."""
        return Layer(int(self) + 1)

    def necromance(self) -> "Layer":
        """Remove the effect of the dead key if any."""
        if self == Layer.ODK:
            return Layer.BASE
        elif self == Layer.ODK_SHIFT:
            return Layer.SHIFT
        return self


@dataclass
class DeadKeyDescr:
    char: str
    name: str
    base: str
    alt: str
    alt_space: str
    alt_self: str


DEAD_KEYS = [DeadKeyDescr(**data) for data in load_data("dead_keys")]

ODK_ID = "**"  # must match the value in dead_keys.yaml
LAYER_KEYS = [
    "- Digits",
    "ae01",
    "ae02",
    "ae03",
    "ae04",
    "ae05",
    "ae06",
    "ae07",
    "ae08",
    "ae09",
    "ae10",
    "- Letters, first row",
    "ad01",
    "ad02",
    "ad03",
    "ad04",
    "ad05",
    "ad06",
    "ad07",
    "ad08",
    "ad09",
    "ad10",
    "- Letters, second row",
    "ac01",
    "ac02",
    "ac03",
    "ac04",
    "ac05",
    "ac06",
    "ac07",
    "ac08",
    "ac09",
    "ac10",
    "- Letters, third row",
    "ab01",
    "ab02",
    "ab03",
    "ab04",
    "ab05",
    "ab06",
    "ab07",
    "ab08",
    "ab09",
    "ab10",
    "- Pinky keys",
    "ae11",
    "ae12",
    "ae13",
    "ad11",
    "ad12",
    "ac11",
    "ab11",
    "tlde",
    "bksl",
    "lsgt",
    "- Space bar",
    "spce",
]



#!/usr/bin/env python3

import json
from contextlib import contextmanager
from importlib import metadata
from pathlib import Path
from typing import Iterator, List, Literal, Union

import click

from .help import create_layout, user_guide
from .layout import KeyboardLayout, load_layout
from .server import keyboard_server


@click.group()
def cli() -> None: ...


def pretty_json(layout: KeyboardLayout, output_path: Path) -> None:
    """Pretty-print the JSON layout.

    Parameters
    ----------
    layout : KeyboardLayout
        The layout to be exported.
    output_path : Path
        The output file path.
    """
    text = (
        json.dumps(layout.json, indent=2, ensure_ascii=False)
        .replace("\n      ", " ")
        .replace("\n    ]", " ]")
        .replace("\n    }", " }")
    )
    output_path.write_text(text, encoding="utf8")


def build_all(layout: KeyboardLayout, output_dir_path: Path) -> None:
    """Generate all layout output files.

    Parameters
    ----------
    layout : KeyboardLayout
        The layout to process.
    output_dir_path : Path
        The output directory.
    msklc_dir : Path
        The MSKLC installation directory.
    """

    @contextmanager
    def file_creation_context(ext: str = "") -> Iterator[Path]:
        """Generate an output file path for extension EXT, return it and finally echo info."""
        path = output_dir_path / (layout.meta["fileName"] + ext)
        yield path
        click.echo(f"... {path}")

    if not output_dir_path.exists():
        output_dir_path.mkdir(parents=True)

    # AHK driver
    with file_creation_context(".ahk") as ahk_path:
        with ahk_path.open("w", encoding="utf-8", newline="\n") as file:
            file.write("\uFEFF")  # AHK scripts require a BOM
            file.write(layout.ahk)

    # Windows driver
    with file_creation_context(".klc") as klc_path:
        with klc_path.open("w", encoding="utf-16le", newline="\r\n") as file:
            try:
                file.write(layout.klc)
            except ValueError as err:
                print(err)

    # macOS driver
    with file_creation_context(".keylayout") as osx_path:
        with osx_path.open("w", encoding="utf-8", newline="\n") as file:
            file.write(layout.keylayout)

    # Linux driver, user-space
    with file_creation_context(".xkb_keymap") as xkb_path:
        with xkb_path.open("w", encoding="utf-8", newline="\n") as file:
            file.write(layout.xkb_keymap)

    # Linux driver, root
    with file_creation_context(".xkb_symbols") as xkb_custom_path:
        with xkb_custom_path.open("w", encoding="utf-8", newline="\n") as file:
            file.write(layout.xkb_symbols.replace("//#", "//"))

    # JSON data
    with file_creation_context(".json") as json_path:
        pretty_json(layout, json_path)

    # SVG data
    with file_creation_context(".svg") as svg_path:
        layout.svg.write(svg_path, encoding="utf-8", xml_declaration=True)


@cli.command()
@click.argument(
    "layout_descriptors",
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    "--out",
    default="all",
    type=click.Path(),
    help="Keyboard drivers to generate.",
)
@click.option(
    "--angle-mod/--no-angle-mod",
    default=False,
    help="Apply Angle-Mod (which is a [ZXCVB] permutation with the LSGT key (a.k.a. ISO key))",
)
@click.option(
    "--qwerty-shortcuts",
    default=False,
    is_flag=True,
    help="Keep shortcuts at their qwerty location",
)
def build(
    layout_descriptors: List[Path],
    out: Union[Path, Literal["all"]],
    angle_mod: bool,
    qwerty_shortcuts: bool,
) -> None:
    """Convert TOML/YAML descriptions into OS-specific keyboard drivers."""

    for input_file in layout_descriptors:
        layout = KeyboardLayout(load_layout(input_file), angle_mod, qwerty_shortcuts)

        # default: build all in the `dist` subdirectory
        if out == "all":
            build_all(layout, Path("dist"))
            continue

        # quick output: reuse the input name and change the file extension
        if out in ["keylayout", "klc", "xkb_keymap", "xkb_symbols", "svg"]:
            output_file = input_file.with_suffix(f".{out}")
        else:
            output_file = Path(out)

        # detailed output
        if output_file.suffix == ".ahk":
            with output_file.open("w", encoding="utf-8", newline="\n") as file:
                file.write("\uFEFF")  # AHK scripts require a BOM
                file.write(layout.ahk)

        elif output_file.suffix == ".klc":
            with output_file.open("w", encoding="utf-16le", newline="\r\n") as file:
                try:
                    file.write(layout.klc)
                except ValueError as err:
                    print(err)

        elif output_file.suffix == ".keylayout":
            with output_file.open("w", encoding="utf-8", newline="\n") as file:
                file.write(layout.keylayout)

        elif output_file.suffix == ".xkb_keymap":
            with output_file.open("w", encoding="utf-8", newline="\n") as file:
                file.write(layout.xkb_keymap)

        elif output_file.suffix == ".xkb_symbols":
            with output_file.open("w", encoding="utf-8", newline="\n") as file:
                file.write(layout.xkb_symbols.replace("//#", "//"))

        elif output_file.suffix == ".json":
            pretty_json(layout, output_file)

        elif output_file.suffix == ".svg":
            layout.svg.write(output_file, encoding="utf-8", xml_declaration=True)

        else:
            click.echo("Unsupported output format.", err=True)
            return

        # successfully converted, display file name
        click.echo(f"... {output_file}")


# TODO: Provide geometry choices
@cli.command()
@click.argument("output_file", nargs=1, type=click.Path(exists=False, path_type=Path))
@click.option("--geometry", default="ISO", help="Specify keyboard geometry.")
@click.option("--altgr/--no-altgr", default=False, help="Set an AltGr layer.")
@click.option("--1dk/--no-1dk", "odk", default=False, help="Set a custom dead key.")
def new(output_file: Path, geometry: str, altgr: bool, odk: bool) -> None:
    """Create a new TOML layout description."""
    create_layout(output_file, geometry, altgr, odk)
    click.echo(f"... {output_file}")


@cli.command()
@click.argument("filepath", nargs=1, type=click.Path(exists=True, path_type=Path))
@click.option(
    "--angle-mod/--no-angle-mod",
    default=False,
    help="Apply Angle-Mod (which is a [ZXCVB] permutation with the LSGT key (a.k.a. ISO key))",
)
def watch(filepath: Path) -> None:
    """Watch a layout description file and display it in a web browser."""
    keyboard_server(filepath)


@cli.command()
def guide() -> None:
    """Show user guide and exit."""
    click.echo(user_guide())


@cli.command()
def version() -> None:
    """Show version number and exit."""
    click.echo(f"kalamine { metadata.version('kalamine') }")


if __name__ == "__main__":
    cli()



import ctypes
import os
import subprocess
import sys
import winreg
from pathlib import Path
from shutil import move, rmtree
from stat import S_IREAD, S_IWUSR

from progress.bar import ChargingBar

from .help import dummy_layout
from .layout import KeyboardLayout


class MsklcManager:
    def __init__(
        self,
        layout: "KeyboardLayout",
        msklc_dir: Path,
        working_dir: Path = Path(os.getcwd()),
        verbose: bool = False,
    ) -> None:
        self._layout = layout
        self._msklc_dir = msklc_dir
        self._verbose = verbose
        self._working_dir = working_dir
        self._progress = ChargingBar(
            f"Creating MSKLC driver for `{layout.meta['name']}`", max=14
        )

    def create_c_files(self):
        """Call kbdutool on the KLC descriptor to generate C files."""

        kbdutool = self._msklc_dir / Path("bin/i386/kbdutool.exe")
        cur = os.getcwd()
        os.chdir(self._working_dir)
        ret = subprocess.run(
            [kbdutool, "-u", "-s", f"{self._layout.meta['name8']}.klc"],
            capture_output=not self._verbose,
        )
        os.chdir(cur)
        ret.check_returncode()

    def _is_already_installed(self) -> bool:
        """Check if the keyboard driver is already installed,
        which would cause MSKLC to launch the GUI instead of creating the installer."""

        # check if the DLL is present
        sys32 = Path(os.environ["WINDIR"]) / Path("System32")
        sysWow = Path(os.environ["WINDIR"]) / Path("SysWOW64")
        dll_name = f'{self._layout.meta["name8"]}.dll'
        dll_exists = (sys32 / dll_name).exists() or (sysWow / Path(dll_name)).exists()

        if dll_exists:
            print(f"Error: {dll_name} is already installed")
            return True

        if sys.platform != "win32":  # let mypy know this is win32-specific
            return False
        # check if the registry still has it
        # that can happen after a botch uninstall of the driver
        kbd_layouts_handle = winreg.OpenKeyEx(
            winreg.HKEY_LOCAL_MACHINE,
            "SYSTEM\\CurrentControlSet\\Control\\Keyboard Layouts",
        )
        # [0] is the number of sub keys
        for i in range(0, winreg.QueryInfoKey(kbd_layouts_handle)[0]):
            sub_key = winreg.EnumKey(kbd_layouts_handle, i)
            sub_handle = winreg.OpenKey(kbd_layouts_handle, sub_key)
            layout_file = winreg.QueryValueEx(sub_handle, "Layout File")[0]
            if layout_file == dll_name:
                print(
                    f"Error: The registry still have reference to `{dll_name}` in"
                    f"`HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Keyboard Layouts\\{sub_key}`"
                )
                return True
        return False

    def _create_dummy_layout(self) -> str:
        return dummy_layout(
            self._layout.geometry,
            self._layout.has_altgr,
            self._layout.has_1dk,
            self._layout.meta,
        ).klc

    def build_msklc_installer(self) -> bool:
        def installer_exists(installer: Path) -> bool:
            return (
                installer.exists()
                and installer.is_dir()
                and (installer / Path("setup.exe")).exists()
                and (installer / Path("amd64")).exists()
                and (installer / Path("i386")).exists()
                and (installer / Path("ia64")).exists()
                and (installer / Path("wow64")).exists()
            )

        name8 = self._layout.meta["name8"]
        installer_dir = self._working_dir / Path(name8)
        if installer_exists(installer_dir):
            self._progress.next(4)
            return True

        if self._is_already_installed():
            self._progress.finish()
            print(
                "Error: layout already installed and "
                "installer package not found in the current directory.\n"
                "Either uninstall the layout manually, or put the installer "
                f"folder in the current directory: ({self._working_dir})"
            )
            return False
        self._progress.message = "Creating installer package"
        self._progress.next()
        # Create a dummy klc file to generate the installer.
        # The file must have a correct name to be reflected in the installer.
        dummy_klc = self._create_dummy_layout()
        klc_file = Path(self._working_dir) / Path(f"{name8}.klc")
        with klc_file.open("w", encoding="utf-16le", newline="\r\n") as file:
            file.write(dummy_klc)

        self._progress.next()
        msklc = self._msklc_dir / Path("MSKLC.exe")
        result = subprocess.run(
            [msklc, klc_file, "-build"], capture_output=not self._verbose, text=True
        )

        self._progress.next()
        # move the installer from "My Documents" to current dir
        if sys.platform != "win32":  # let mypy know this is win32-specific
            return False

        CSIDL_PERSONAL = 5  # My Documents
        SHGFP_TYPE_CURRENT = 0  # Get current, not default value
        buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(
            None, CSIDL_PERSONAL, None, SHGFP_TYPE_CURRENT, buf
        )
        my_docs = Path(buf.value)
        installer = my_docs / Path(name8)

        self._progress.next()
        if not installer_exists(installer):
            self._progress.finish()
            print(f"MSKLC Exit code: {result.returncode}")
            print(result.stdout)
            print(result.stderr)
            print("Error: installer was not created.")
            return False

        move(str(installer), str(self._working_dir / Path(name8)))
        return True

    def build_msklc_dll(self) -> bool:
        self._progress.next()
        name8 = self._layout.meta["name8"]
        prev = os.getcwd()
        os.chdir(self._working_dir)
        INST_DIR = self._working_dir / Path(name8)
        dll_dirs = ["i386", "amd64", "ia64", "wow64"]
        for dll_dir in dll_dirs:
            full_dir = INST_DIR / Path(dll_dir)
            if not full_dir.exists():
                raise Exception(f"{full_dir} doesn't exist")
            else:
                rmtree(full_dir)
                os.mkdir(full_dir)

        self._progress.next()
        # create correct klc
        klc_file = self._working_dir / Path(f"{name8}.klc")
        with klc_file.open("w", encoding="utf-16le", newline="\r\n") as file:
            try:
                file.write(self._layout.klc)
            except ValueError as err:
                print(f"ERROR: {err}")
                return False

        self._progress.next()
        self.create_c_files()

        self._progress.next()
        rc_file = klc_file.with_suffix(".RC")
        with rc_file.open("w", encoding="utf-16le", newline="\r\n") as file:
            file.write(self._layout.klc_rc)

        self._progress.next()
        c_file = klc_file.with_suffix(".C")
        with c_file.open("w", encoding="utf-16le", newline="\r\n") as file:
            file.write(self._layout.klc_c)

        c_files = [".C", ".RC", ".H", ".DEF"]

        self._progress.next()
        # Make files read-only to prevent MSKLC from overwriting them.
        for suffix in c_files:
            os.chmod(klc_file.with_suffix(suffix), S_IREAD)

        # build correct DLLs

        kbdutool = self._msklc_dir / Path("bin/i386/kbdutool.exe")
        dll = klc_file.with_suffix(".dll")
        self._progress.message = "Creating driver DLLs"
        for arch_flag, arch in [
            ("-x", "i386"),
            ("-m", "amd64"),
            ("-i", "ia64"),
            ("-o", "wow64"),
        ]:
            self._progress.next()
            result = subprocess.run(
                [kbdutool, "-u", arch_flag, klc_file],
                text=True,
                capture_output=not self._verbose,
            )
            if result.returncode == 0:
                move(str(dll), str(INST_DIR / Path(arch)))
            else:
                # Restore write permission
                for suffix in c_files:
                    os.chmod(klc_file.with_suffix(suffix), S_IWUSR)
                self._progress.finish()
                print(f"Error while creating DLL for arch {arch}:")
                print(result.stdout)
                print(result.stderr)
                return False

        # Restore write permission
        for suffix in c_files:
            os.chmod(klc_file.with_suffix(suffix), S_IWUSR)
        os.chdir(prev)
        self._progress.finish()
        return True



"""
Helper to list, add and remove keyboard layouts from XKB config.
This MUST remain dependency-free in order to be usable as a standalone installer.
"""

import datetime
import re
import sys
import traceback
from os import environ
from pathlib import Path
from textwrap import dedent
from typing import Dict, ItemsView, Optional
from xml.etree import ElementTree as ET

from .layout import KeyboardLayout


def xdg_config_home() -> Path:
    xdg_config = environ.get("XDG_CONFIG_HOME")
    if xdg_config:
        return Path(xdg_config)
    return Path.home() / ".config"


def wayland_running() -> bool:
    xdg_session = environ.get("XDG_SESSION_TYPE")
    if xdg_session:
        return xdg_session.startswith("wayland")
    return False


XKB_HOME = xdg_config_home() / "xkb"
XKB_ROOT = Path(environ.get("XKB_CONFIG_ROOT") or "/usr/share/X11/xkb/")

WAYLAND = wayland_running()

KALAMINE_MARK = f"Generated by kalamine on {datetime.date.today().isoformat()}"
XKB_RULES_HEADER = f"""\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE xkbConfigRegistry SYSTEM "xkb.dtd">
<!-- {KALAMINE_MARK} -->
"""

LayoutName = str
LocaleName = str
KbdVariant = Dict[LayoutName, Optional[KeyboardLayout]]
KbdIndex = Dict[LocaleName, KbdVariant]
XmlIndex = Dict[LocaleName, Dict[LayoutName, str]]


class XKBManager:
    """Wrapper to list/add/remove keyboard drivers to XKB."""

    def __init__(self, root: bool = False) -> None:
        self._as_root = root
        self._rootdir = XKB_ROOT if root else XKB_HOME
        self._index: KbdIndex = {}

    @property
    def index(self) -> ItemsView[LocaleName, KbdVariant]:
        return self._index.items()

    @property
    def path(self) -> Path:
        return self._rootdir

    def add(self, layout: KeyboardLayout) -> None:
        locale = layout.meta["locale"]
        variant = layout.meta["variant"]
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = layout

    def remove(self, locale: str, variant: str) -> None:
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = None

    def update(self) -> None:
        update_rules(self._rootdir, self._index)  # XKB/rules/{base,evdev}.xml
        update_symbols(self._rootdir, self._index)  # XKB/symbols/{locales}
        self._index = {}

    def clean(self) -> None:
        """Drop the obsolete 'type' attributes Kalamine used to add."""
        for filename in ["base.xml", "evdev.xml"]:
            filepath = self._rootdir / "rules" / filename
            if not filepath.exists():
                continue
            tree = ET.parse(str(filepath))
            for variant in tree.findall(".//variant[@type]"):
                variant.attrib.pop("type")

    def list(self, mask: str = "") -> XmlIndex:
        layouts = list_rules(self._rootdir, mask)
        return list_symbols(self._rootdir, layouts)

    def list_all(self, mask: str = "") -> XmlIndex:
        return list_rules(self._rootdir, mask)

    def has_custom_symbols(self) -> bool:
        """Check if there is a usable xkb/symbols/custom file."""

        custom_path = self._rootdir / "symbols" / "custom"
        if not custom_path.exists():
            return False

        for filename in ["base.xml", "evdev.xml"]:
            filepath = self._rootdir / "rules" / filename
            if not filepath.exists():
                continue
            tree = ET.parse(str(filepath))
            if tree.find('.//layout/configItem/name[.="custom"]'):
                return True

        return False

    def ensure_xkb_config_is_ready(self) -> None:
        """Ensure there is an XKB configuration in user-space."""
        # See xkblayout.py for a more extensive version of this feature:
        # https://gitlab.freedesktop.org/whot/xkblayout

        if self._as_root:
            return

        # ensure all expected directories exist (don't care about 'geometry')
        XKB_HOME.mkdir(exist_ok=True)
        for subdir in ["compat", "keycodes", "rules", "symbols", "types"]:
            (XKB_HOME / subdir).mkdir(exist_ok=True)

        # ensure there are XKB rules
        # (new locales and symbols will be added by XKBManager)
        for ruleset in ["evdev"]:  # add 'base', too?
            # xkb/rules/evdev
            rules = XKB_HOME / "rules" / ruleset
            if not rules.exists():
                rules.write_text(
                    dedent(
                        f"""\
                        // {KALAMINE_MARK}
                        // Include the system '{ruleset}' file
                        ! include %S/{ruleset}
                        """
                    )
                )
            # xkb/rules/evdev.xml
            xmlpath = XKB_HOME / "rules" / f"{ruleset}.xml"
            if not xmlpath.exists():
                xmlpath.write_text(
                    XKB_RULES_HEADER
                    + dedent(
                        """\
                        <xkbConfigRegistry version="1.1">
                            <layoutList/>
                        </xkbConfigRegistry>
                        """
                    )
                )


""" On GNU/Linux, keyboard layouts must be installed in /usr/share/X11/xkb. To
    be able to revert a layout installation, Kalamine marks layouts like this:

    - XKB/symbols/[locale]: layout definitions
        // KALAMINE::[NAME]::BEGIN
        xkb_symbols "[name]" { ... }
        // KALAMINE::[NAME]::END

    Earlier versions of XKalamine used to mark index files as well but recent
    versions of Gnome do not support the custom `type` attribute any more, which
    must be removed:

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="kalamine">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Even worse, the Lafayette project has released a first installer before
    the XKalamine installer was developed, so we have to handle this situation
    too:

    - XKB/symbols/[locale]: layout definitions
        // LAFAYETTE::BEGIN
        xkb_symbols "lafayette"   { ... }
        xkb_symbols "lafayette42" { ... }
        // LAFAYETTE::END

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="lafayette">
            <configItem>
                <name>lafayette</name>
                <description>French (Lafayette)</description>
            </configItem>
        </variant>
        <variant type="lafayette">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Consequence: these two Lafayette layouts must be uninstalled together.
    Because of the way they are grouped in symbols/fr, it is impossible to
    remove one without removing the other.
"""


def clean_legacy_lafayette() -> None:
    return


###############################################################################
# Helpers: XKB/symbols
#


LEGACY_MARK = {"begin": "// LAFAYETTE::BEGIN\n", "end": "// LAFAYETTE::END\n"}


def get_symbol_mark(name: str) -> Dict[str, str]:
    return {
        "begin": "// KALAMINE::" + name.upper() + "::BEGIN\n",
        "end": "// KALAMINE::" + name.upper() + "::END\n",
    }


def is_new_symbol_mark(line: str) -> Optional[str]:
    if not line.endswith("::BEGIN\n"):
        return None

    if line.startswith("// KALAMINE::"):
        return line[13:-8].lower()  # XXX Kalamine expects lowercase names

    return "lafayette"  # line.startswith("// LAFAYETTE::"):  # obsolete marker


def update_symbols_locale(path: Path, named_layouts: KbdVariant) -> None:
    """Update Kalamine layouts in an xkb/symbols/[locale] file."""

    text = ""
    modified_text = False
    with path.open("r+", encoding="utf-8") as symbols:
        # look for Kalamine layouts to be updated or removed
        between_marks = False
        closing_mark = ""
        for line in symbols:
            name = is_new_symbol_mark(line)
            if name:
                if name in named_layouts.keys():
                    closing_mark = line[:-6] + "END\n"
                    modified_text = True
                    between_marks = True
                    text = text.rstrip()
                else:
                    text += line
            elif line.endswith("::END\n"):
                if between_marks and line.startswith(closing_mark):
                    between_marks = False
                    closing_mark = ""
                else:
                    text += line
            elif not between_marks:
                text += line

        # clear previous Kalamine layouts if needed
        if modified_text:
            symbols.seek(0)
            symbols.write(text.rstrip() + "\n")
            symbols.truncate()

        # add new Kalamine layouts
        locale = path.name
        for name, layout in named_layouts.items():
            if layout is None:
                print(f"      - {locale}/{name}")
            else:
                print(f"      + {locale}/{name}")
                mark = get_symbol_mark(name)
                symbols.write("\n")
                symbols.write(mark["begin"])
                symbols.write(
                    re.sub(  # drop lines starting with '//#'
                        r"^//#.*\n", "", layout.xkb_symbols, flags=re.MULTILINE
                    ).rstrip()
                    + "\n"
                )
                symbols.write(mark["end"])

        symbols.close()


def update_symbols(xkb_root: Path, kbd_index: KbdIndex) -> None:
    """Update Kalamine layouts in all xkb/symbols files."""

    for locale, named_layouts in kbd_index.items():
        path = xkb_root / "symbols" / locale
        if not path.exists():
            with path.open("w") as file:
                file.write("// {KALAMINE_MARK}")
                file.close()

        try:
            print(f"... {path}")
            update_symbols_locale(path, named_layouts)

        except Exception as exc:
            exit_FileNotWritable(exc, path)


def list_symbols(xkb_root: Path, xml_index: XmlIndex) -> XmlIndex:
    """Filter input layouts: only keep the ones defined with Kalamine."""

    filtered_index: XmlIndex = {}
    for locale, variants in sorted(xml_index.items()):
        path = xkb_root / "symbols" / locale
        if not path.exists():
            continue

        with open(path, "r", encoding="utf-8") as symbols:
            for line in symbols:
                name = is_new_symbol_mark(line)
                if name is None:
                    continue
                if name in variants.keys():
                    if locale not in filtered_index:
                        filtered_index[locale] = {}
                    filtered_index[locale][name] = variants[name]

    return filtered_index


###############################################################################
# Helpers: XKB/rules
#


def get_rules_variant_list(
    tree: ET.ElementTree, locale: LocaleName
) -> Optional[ET.Element]:
    """Find the <layout/variantList> item matching the locale."""

    query = f'.//layout/configItem/name[.="{locale}"]/../../variantList'

    if tree.find(query) is None:  # create the locale if needed
        layout_list = tree.find(".//layoutList")
        if layout_list is None:
            raise Exception
        layout = ET.SubElement(layout_list, "layout")
        config = ET.SubElement(layout, "configItem")
        ET.SubElement(config, "name").text = locale
        ET.SubElement(layout, "variantList")

    return tree.find(query)


def remove_rules_variant(variant_list: ET.Element, name: str) -> None:
    """Remove a named <variant> item from <layout/variantList>."""

    for variant in variant_list.findall(f'variant/configItem/name[.="{name}"]/../..'):
        variant_list.remove(variant)


def add_rules_variant(variant_list: ET.Element, name: str, description: str) -> None:
    """Add a <variant/configItem/{name,description> item to <layout/variantList>."""

    variant = ET.SubElement(variant_list, "variant")
    config = ET.SubElement(variant, "configItem")
    ET.SubElement(config, "name").text = name
    ET.SubElement(config, "description").text = description


def update_rules(xkb_root: Path, kbd_index: KbdIndex) -> None:
    """Update references in XKB/rules/{base,evdev}.xml."""

    for filename in ["base.xml", "evdev.xml"]:
        filepath = xkb_root / "rules" / filename
        if not filepath.exists():
            continue

        try:
            tree = ET.parse(filepath)

            for locale, named_layouts in kbd_index.items():
                vlist = get_rules_variant_list(tree, locale)
                if vlist is None:
                    exit(f"Error: unexpected xml format in {filepath}.")

                for name, layout in named_layouts.items():
                    remove_rules_variant(vlist, name)
                    if layout is not None:
                        description = layout.meta["description"]
                        add_rules_variant(vlist, name, description)

            if hasattr(ET, "indent"):  # Python 3.9+
                ET.indent(tree)

            with filepath.open("w") as file:
                file.write(XKB_RULES_HEADER)
                file.write(ET.tostring(tree.getroot(), encoding="unicode"))
                file.close()

            print(f"... {filepath}")

        except Exception as exc:
            exit_FileNotWritable(exc, filepath)


def list_rules(xkb_root: Path, mask: str = "*") -> XmlIndex:
    """List all matching XKB layouts."""

    if mask in ("", "*"):
        locale_mask = "*"
        variant_mask = "*"
    else:
        m = mask.split("/")
        if len(m) == 2:
            locale_mask, variant_mask = m
        else:
            locale_mask = mask
            variant_mask = "*"

    xml_index: XmlIndex = {}
    for filename in ["base.xml", "evdev.xml"]:
        filepath = xkb_root / "rules" / filename
        if not filepath.exists():
            continue

        tree = ET.parse(filepath)
        locales = [str(name.text) for name in tree.findall(".//layout/configItem/name")]
        for locale in locales:
            for variant in tree.findall(
                f'.//layout/configItem/name[.="{locale}"]/../../variantList/variant'
            ):
                name = variant.find("configItem/name")
                desc = variant.find("configItem/description")
                if name is None or name.text is None or desc is None:
                    continue

                if locale_mask in ("*", locale) and variant_mask in ("*", name.text):
                    if locale not in xml_index:
                        xml_index[(locale)] = {}
                    xml_index[locale][name.text] = str(desc.text)

    return xml_index


###############################################################################
# Exception Handling (there must be a better way...)
#


def exit_FileNotWritable(exception: Exception, path: Path) -> None:
    if isinstance(exception, PermissionError):  # noqa: F821
        raise exception
    if isinstance(exception, IOError):
        print("")
        sys.exit(f"Error: could not write to file {path}.")
    else:
        print("")
        sys.exit(f"Error: {exception}.\n{traceback.format_exc()}")



from kalamine import KeyboardLayout

from .util import get_layout_dict


def load_layout(filename: str, angle_mod: bool = False) -> KeyboardLayout:
    return KeyboardLayout(get_layout_dict(filename), angle_mod)


def test_ansi():
    layout = load_layout("ansi")
    assert layout.layers[0]["ad01"] == "q"
    assert layout.layers[1]["ad01"] == "Q"
    assert layout.layers[0]["tlde"] == "`"
    assert layout.layers[1]["tlde"] == "~"
    assert not layout.has_altgr
    assert not layout.has_1dk
    assert "**" not in layout.dead_keys

    # ensure angle mod is NOT applied
    layout = load_layout("ansi", angle_mod=True)
    assert layout.layers[0]["ab01"] == "z"
    assert layout.layers[1]["ab01"] == "Z"


def test_prog():  # AltGr + dead keys
    layout = load_layout("prog")
    assert layout.layers[0]["ad01"] == "q"
    assert layout.layers[1]["ad01"] == "Q"
    assert layout.layers[0]["tlde"] == "`"
    assert layout.layers[1]["tlde"] == "~"
    assert layout.layers[4]["tlde"] == "*`"
    assert layout.layers[5]["tlde"] == "*~"
    assert layout.has_altgr
    assert not layout.has_1dk
    assert "**" not in layout.dead_keys
    assert len(layout.dead_keys["*`"]) == 18
    assert len(layout.dead_keys["*~"]) == 21


def test_intl():  # 1dk + dead keys
    layout = load_layout("intl")
    assert layout.layers[0]["ad01"] == "q"
    assert layout.layers[1]["ad01"] == "Q"
    assert layout.layers[0]["tlde"] == "*`"
    assert layout.layers[1]["tlde"] == "*~"
    assert not layout.has_altgr
    assert layout.has_1dk
    assert "**" in layout.dead_keys

    assert len(layout.dead_keys) == 5
    assert "**" in layout.dead_keys
    assert "*`" in layout.dead_keys
    assert "*^" in layout.dead_keys
    assert "*¨" in layout.dead_keys
    assert "*~" in layout.dead_keys
    assert len(layout.dead_keys["**"]) == 15
    assert len(layout.dead_keys["*`"]) == 18
    assert len(layout.dead_keys["*^"]) == 43
    assert len(layout.dead_keys["*¨"]) == 21
    assert len(layout.dead_keys["*~"]) == 21

    # ensure the 1dk parser does not accumulate values from a previous run
    layout = load_layout("intl")
    assert len(layout.dead_keys["*`"]) == 18
    assert len(layout.dead_keys["*~"]) == 21

    assert len(layout.dead_keys) == 5
    assert "**" in layout.dead_keys
    assert "*`" in layout.dead_keys
    assert "*^" in layout.dead_keys
    assert "*¨" in layout.dead_keys
    assert "*~" in layout.dead_keys
    assert len(layout.dead_keys["**"]) == 15
    assert len(layout.dead_keys["*`"]) == 18
    assert len(layout.dead_keys["*^"]) == 43
    assert len(layout.dead_keys["*¨"]) == 21
    assert len(layout.dead_keys["*~"]) == 21

    # ensure angle mod is working correctly
    layout = load_layout("intl", angle_mod=True)
    assert layout.layers[0]["lsgt"] == "z"
    assert layout.layers[1]["lsgt"] == "Z"
    assert layout.layers[0]["ab01"] == "x"
    assert layout.layers[1]["ab01"] == "X"



#!/usr/bin/env python3

import platform
import sys
from pathlib import Path
from typing import List

import click

from .layout import KeyboardLayout, load_layout
from .msklc_manager import MsklcManager


@click.group()
def cli() -> None: ...


DEFAULT_MSKLC_DIR = "C:\\Program Files (x86)\\Microsoft Keyboard Layout Creator 1.4\\"


@cli.command()
@click.argument(
    "layout_descriptors",
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    "--angle-mod/--no-angle-mod",
    default=False,
    help="Apply Angle-Mod (which is a [ZXCVB] permutation with the LSGT key (a.k.a. ISO key))",
)
@click.option(
    "--msklc",
    default=DEFAULT_MSKLC_DIR,
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Directory where MSKLC is installed",
)
@click.option(
    "--qwerty-shortcuts",
    default=False,
    is_flag=True,
    help="Keep shortcuts at their qwerty location",
)
@click.option("--verbose", "-v", is_flag=True, help="Verbose mode")
def build(
    layout_descriptors: List[Path],
    angle_mod: bool,
    msklc: Path,
    qwerty_shortcuts: bool,
    verbose: bool,
) -> None:
    """Convert TOML/YAML descriptions into Windows MSKLC keyboard drivers."""

    if platform.system() != "Windows":
        sys.exit("This command is only compatible with Windows, sorry.")

    for input_file in layout_descriptors:
        layout = KeyboardLayout(load_layout(input_file), angle_mod, qwerty_shortcuts)
        msklc_mgr = MsklcManager(layout, msklc, verbose=verbose)
        if msklc_mgr.build_msklc_installer():
            if msklc_mgr.build_msklc_dll():
                output_dir = f'{msklc_mgr._working_dir}\\{layout.meta["name8"]}\\'
                click.echo(
                    "MSKLC drivers successfully built.\n"
                    f"Execute `{output_dir}setup.exe` to install.\n"
                    "Log out and log back in to apply the changes."
                )


@cli.command()
@click.argument(
    "input_file",
    nargs=1,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option("--verbose", "-v", is_flag=True, help="Verbose mode")
def dummy(input_file: Path, verbose: bool) -> None:
    """Dump a dummy TOML layout descriptor."""

    input_layout = KeyboardLayout(load_layout(input_file))
    msklc_mgr = MsklcManager(input_layout, Path(DEFAULT_MSKLC_DIR), verbose=verbose)
    print(msklc_mgr._create_dummy_layout())


if __name__ == "__main__":
    cli()



"""Some util functions for tests."""

from pathlib import Path
from typing import Dict

import tomli


def get_layout_dict(filename: str) -> Dict:
    """Return the layout directory path."""

    file_path = Path(__file__).parent.parent / f"layouts/{filename}.toml"
    with file_path.open(mode="rb") as file:
        return tomli.load(file)



import copy
import datetime
import pkgutil
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Set, Type, TypeVar
from xml.etree import ElementTree as ET

import click
import tomli
import yaml

from .template import (
    ahk_keymap,
    ahk_shortcuts,
    c_deadkeys,
    c_dk_index,
    c_keymap,
    klc_deadkeys,
    klc_dk_index,
    klc_keymap,
    osx_actions,
    osx_keymap,
    osx_terminators,
    web_deadkeys,
    web_keymap,
    xkb_keymap,
)
from .utils import (
    DEAD_KEYS,
    LAYER_KEYS,
    ODK_ID,
    Layer,
    lines_to_text,
    load_data,
    text_to_lines,
)

###
# Helpers
#


def get_langid(locale: str) -> str:
    locale_codes = load_data("win_locales")
    if locale not in locale_codes:
        raise ValueError(f"`{locale}` is not a valid locale")
    return locale_codes[locale]


def upper_key(letter: Optional[str], blank_if_obvious: bool = True) -> str:
    """This is used for presentation purposes: in a key, the upper character
    becomes blank if it's an obvious uppercase version of the base character."""

    if letter is None:
        return " "

    custom_alpha = {
        "\u00df": "\u1e9e",  # ß ẞ
        "\u007c": "\u00a6",  # | ¦
        "\u003c": "\u2264",  # < ≤
        "\u003e": "\u2265",  # > ≥
        "\u2020": "\u2021",  # † ‡
        "\u2190": "\u21d0",  # ← ⇐
        "\u2191": "\u21d1",  # ↑ ⇑
        "\u2192": "\u21d2",  # → ⇒
        "\u2193": "\u21d3",  # ↓ ⇓
        "\u00b5": " ",  # µ (to avoid getting `Μ` as uppercase)
    }
    if letter in custom_alpha:
        return custom_alpha[letter]

    if len(letter) == 1 and letter.upper() != letter.lower():
        return letter.upper()

    # dead key or non-letter character
    return " " if blank_if_obvious else letter


def substitute_lines(text: str, variable: str, lines: List[str]) -> str:
    prefix = "KALAMINE::"
    exp = re.compile(".*" + prefix + variable + ".*")

    indent = ""
    for line in text.split("\n"):
        m = exp.match(line)
        if m:
            indent = m.group().split(prefix)[0]
            break

    return exp.sub(lines_to_text(lines, indent), text)


def substitute_token(text: str, token: str, value: str) -> str:
    exp = re.compile("\\$\\{" + token + "(=[^\\}]*){0,1}\\}")
    return exp.sub(value, text)


def load_tpl(layout: "KeyboardLayout", ext: str, tpl: str = "base") -> str:
    date = datetime.date.today().isoformat()
    if tpl == "base":
        if layout.has_altgr or ext.startswith(".RC"):
            tpl = "full"
            if layout.has_1dk and ext.startswith(".xkb"):
                tpl = "full_1dk"
    bin = pkgutil.get_data(__package__, f"tpl/{tpl}{ext}")
    if bin is None:
        return ""
    out = bin.decode("utf-8")
    out = substitute_lines(out, "GEOMETRY_base", layout.base)
    out = substitute_lines(out, "GEOMETRY_full", layout.full)
    out = substitute_lines(out, "GEOMETRY_altgr", layout.altgr)
    out = substitute_token(out, "KALAMINE", f"Generated by kalamine on {date}")
    for key, value in layout.meta.items():
        out = substitute_token(out, key, value)
    return out


def load_layout(layout_path: Path) -> Dict:
    """Load the TOML/YAML layout description data (and its ancessor, if any)."""

    def load_descriptor(file_path: Path) -> Dict:
        if file_path.suffix in [".yaml", ".yml"]:
            with file_path.open(encoding="utf-8") as file:
                return yaml.load(file, Loader=yaml.SafeLoader)

        with file_path.open(mode="rb") as dfile:
            return tomli.load(dfile)

    try:
        cfg = load_descriptor(layout_path)
        if "name" not in cfg:
            cfg["name"] = layout_path.stem
        if "extends" in cfg:
            parent_path = layout_path.parent / cfg["extends"]
            ext = load_descriptor(parent_path)
            ext.update(cfg)
            cfg = ext
        return cfg

    except Exception as exc:
        click.echo("File could not be parsed.", err=True)
        click.echo(f"Error: {exc}.", err=True)
        sys.exit(1)


###
# Constants
#


# fmt: off
@dataclass
class MetaDescr:
    name:        str = "custom"
    name8:       str = "custom"
    variant:     str = "custom"
    fileName:    str = "custom"
    locale:      str = "us"
    geometry:    str = "ISO"
    description: str = ""
    author:      str = "nobody"
    license:     str = ""
    version:     str = "0.0.1"


@dataclass
class SpacebarDescr:
    shift:       str = " "
    altgr:       str = " "
    altgt_shift: str = " "
    odk:         str = "'"
    odk_shift:   str = "'"
# fmt: on


CONFIG = {
    "author": "nobody",
    "license": "WTFPL - Do What The Fuck You Want Public License",
    "geometry": "ISO",
}

SPACEBAR = {
    "shift": " ",
    "altgr": " ",
    "altgr_shift": " ",
    "1dk": "'",
    "1dk_shift": "'",
}


@dataclass
class RowDescr:
    offset: int
    keys: List[str]


T = TypeVar("T", bound="GeometryDescr")


@dataclass
class GeometryDescr:
    template: str
    rows: List[RowDescr]

    @classmethod
    def from_dict(cls: Type[T], src: Dict) -> T:
        return cls(
            template=src["template"], rows=[RowDescr(**row) for row in src["rows"]]
        )


GEOMETRY = {
    key: GeometryDescr.from_dict(val) for key, val in load_data("geometry").items()
}


###
# Main
#


class KeyboardLayout:
    """Lafayette-style keyboard layout: base + 1dk + altgr layers."""

    # self.meta = {key: MetaDescr.from_dict(val) for key, val in geometry_data.items()}

    def __init__(
        self, layout_data: Dict, angle_mod: bool = False, qwerty_shortcuts: bool = False
    ) -> None:
        """Import a keyboard layout to instanciate the object."""

        # initialize a blank layout
        self.layers: Dict[Layer, Dict[str, str]] = {layer: {} for layer in Layer}
        self.dk_set: Set[str] = set()
        self.dead_keys: Dict[str, Dict[str, str]] = {}  # dictionary subset of DEAD_KEYS
        # self.meta = Dict[str, str] = {} # default parameters, hardcoded
        self.meta = CONFIG.copy()  # default parameters, hardcoded
        self.has_altgr = False
        self.has_1dk = False
        self.qwerty_shortcuts = qwerty_shortcuts

        # metadata: self.meta
        for k in layout_data:
            if (
                k != "base"
                and k != "full"
                and k != "altgr"
                and not isinstance(layout_data[k], dict)
            ):
                self.meta[k] = layout_data[k]
        self.meta["name8"] = (
            layout_data["name8"] if "name8" in layout_data else self.meta["name"][0:8]
        )
        self.meta["fileName"] = self.meta["name8"].lower()

        # keyboard layers: self.layers & self.dead_keys
        rows = copy.deepcopy(GEOMETRY[self.meta["geometry"]].rows)

        # Angle Mod permutation
        if angle_mod:
            last_row = rows[3]
            if last_row.keys[0] == "lsgt":
                # should bevome ['ab05', 'lsgt', 'ab01', 'ab02', 'ab03', 'ab04']
                last_row.keys[:6] = [last_row.keys[5]] + last_row.keys[:5]
            else:
                click.echo(
                    "Warning: geometry does not support angle-mod; ignoring the --angle-mod argument"
                )

        if "full" in layout_data:
            full = text_to_lines(layout_data["full"])
            self._parse_template(full, rows, Layer.BASE)
            self._parse_template(full, rows, Layer.ALTGR)
            self.has_altgr = True
        else:
            base = text_to_lines(layout_data["base"])
            self._parse_template(base, rows, Layer.BASE)
            self._parse_template(base, rows, Layer.ODK)
            if "altgr" in layout_data:
                self.has_altgr = True
                self._parse_template(
                    text_to_lines(layout_data["altgr"]), rows, Layer.ALTGR
                )

        # space bar
        spc = SPACEBAR.copy()
        if "spacebar" in layout_data:
            for k in layout_data["spacebar"]:
                spc[k] = layout_data["spacebar"][k]
        self.layers[Layer.BASE]["spce"] = " "
        self.layers[Layer.SHIFT]["spce"] = spc["shift"]
        if True or self.has_1dk:  # XXX self.has_1dk is not defined yet
            self.layers[Layer.ODK]["spce"] = spc["1dk"]
            self.layers[Layer.ODK_SHIFT]["spce"] = (
                spc["shift_1dk"] if "shift_1dk" in spc else spc["1dk"]
            )
        if self.has_altgr:
            self.layers[Layer.ALTGR]["spce"] = spc["altgr"]
            self.layers[Layer.ALTGR_SHIFT]["spce"] = spc["altgr_shift"]

        self._parse_dead_keys(spc)

    def _parse_dead_keys(self, spc: Dict[str, str]) -> None:
        """Build a deadkey dict."""

        def layout_has_char(char: str) -> bool:
            all_layers = [Layer.BASE, Layer.SHIFT]
            if self.has_altgr:
                all_layers += [Layer.ALTGR, Layer.ALTGR_SHIFT]

            for layer_index in all_layers:
                for id in self.layers[layer_index]:
                    if self.layers[layer_index][id] == char:
                        return True
            return False

        all_spaces: List[str] = []
        for space in ["\u0020", "\u00a0", "\u202f"]:
            if layout_has_char(space):
                all_spaces.append(space)

        self.dead_keys = {}
        for dk in DEAD_KEYS:
            id = dk.char
            if id not in self.dk_set:
                continue

            self.dead_keys[id] = {}
            deadkey = self.dead_keys[id]
            deadkey[id] = dk.alt_self

            if id == ODK_ID:
                self.has_1dk = True
                for key_name in LAYER_KEYS:
                    if key_name.startswith("-"):
                        continue
                    for layer in [Layer.ODK_SHIFT, Layer.ODK]:
                        if key_name in self.layers[layer]:
                            deadkey[self.layers[layer.necromance()][key_name]] = (
                                self.layers[layer][key_name]
                            )
                for space in all_spaces:
                    deadkey[space] = spc["1dk"]

            else:
                base = dk.base
                alt = dk.alt
                for i in range(len(base)):
                    if layout_has_char(base[i]):
                        deadkey[base[i]] = alt[i]
                for space in all_spaces:
                    deadkey[space] = dk.alt_space

    def _parse_template(
        self, template: List[str], rows: List[RowDescr], layer_number: Layer
    ) -> None:
        """Extract a keyboard layer from a template."""

        j = 0
        col_offset = 0 if layer_number == Layer.BASE else 2
        for row in rows:
            i = row.offset + col_offset
            keys = row.keys

            base = list(template[2 + j * 3])
            shift = list(template[1 + j * 3])

            for key in keys:
                base_key = ("*" if base[i - 1] == "*" else "") + base[i]
                shift_key = ("*" if shift[i - 1] == "*" else "") + shift[i]

                # in the BASE layer, if the base character is undefined, shift prevails
                if base_key == " ":
                    if layer_number == Layer.BASE:
                        base_key = shift_key.lower()

                # in other layers, if the shift character is undefined, base prevails
                elif shift_key == " ":
                    if layer_number == Layer.ALTGR:
                        shift_key = upper_key(base_key)
                    elif layer_number == Layer.ODK:
                        shift_key = upper_key(base_key)
                        # shift_key = upper_key(base_key, blank_if_obvious=False)

                if base_key != " ":
                    self.layers[layer_number][key] = base_key
                if shift_key != " ":
                    self.layers[layer_number.next()][key] = shift_key

                for dk in DEAD_KEYS:
                    if base_key == dk.char or shift_key == dk.char:
                        self.dk_set.add(dk.char)

                i += 6
            j += 1

    ###
    # Geometry: base, full, altgr
    #

    def _fill_template(
        self, template: List[str], rows: List[RowDescr], layer_number: Layer
    ) -> List[str]:
        """Fill a template with a keyboard layer."""

        if layer_number == Layer.BASE:
            col_offset = 0
            shift_prevails = True
        else:  # AltGr or 1dk
            col_offset = 2
            shift_prevails = False

        j = 0
        for row in rows:
            i = row.offset + col_offset
            keys = row.keys

            base = list(template[2 + j * 3])
            shift = list(template[1 + j * 3])

            for key in keys:
                base_key = " "
                if key in self.layers[layer_number]:
                    base_key = self.layers[layer_number][key]

                shift_key = " "
                if key in self.layers[layer_number.next()]:
                    shift_key = self.layers[layer_number.next()][key]

                dead_base = len(base_key) == 2 and base_key[0] == "*"
                dead_shift = len(shift_key) == 2 and shift_key[0] == "*"

                if shift_prevails:
                    shift[i] = shift_key[-1]
                    if dead_shift:
                        shift[i - 1] = "*"
                    if upper_key(base_key) != shift_key:
                        base[i] = base_key[-1]
                        if dead_base:
                            base[i - 1] = "*"
                else:
                    base[i] = base_key[-1]
                    if dead_base:
                        base[i - 1] = "*"
                    if upper_key(base_key) != shift_key:
                        shift[i] = shift_key[-1]
                        if dead_shift:
                            shift[i - 1] = "*"

                i += 6

            template[2 + j * 3] = "".join(base)
            template[1 + j * 3] = "".join(shift)
            j += 1

        return template

    def _get_geometry(self, layers: Optional[List[Layer]] = None) -> List[str]:
        """`geometry` view of the requested layers."""
        layers = layers or [Layer.BASE]

        rows = GEOMETRY[self.geometry].rows
        template = GEOMETRY[self.geometry].template.split("\n")[:-1]
        for i in layers:
            template = self._fill_template(template, rows, i)
        return template

    @property
    def geometry(self) -> str:
        """ANSI, ISO, ERGO."""
        return self.meta["geometry"].upper()

    @geometry.setter
    def geometry(self, value: str) -> None:
        """ANSI, ISO, ERGO."""
        shape = value.upper()
        if shape not in ["ANSI", "ISO", "ERGO"]:
            shape = "ISO"
        self.meta["geometry"] = shape

    @property
    def base(self) -> List[str]:
        """Base + 1dk layers."""
        return self._get_geometry([Layer.BASE, Layer.ODK])

    @property
    def full(self) -> List[str]:
        """Base + AltGr layers."""
        return self._get_geometry([Layer.BASE, Layer.ALTGR])

    @property
    def altgr(self) -> List[str]:
        """AltGr layer only."""
        return self._get_geometry([Layer.ALTGR])

    ###
    # OS-specific drivers: keylayout, klc, xkb, xkb_patch
    #

    @property
    def keylayout(self) -> str:
        """macOS driver"""
        out = load_tpl(self, ".keylayout")
        for i, layer in enumerate(osx_keymap(self)):
            out = substitute_lines(out, "LAYER_" + str(i), layer)
        out = substitute_lines(out, "ACTIONS", osx_actions(self))
        out = substitute_lines(out, "TERMINATORS", osx_terminators(self))
        return out

    @property
    def ahk(self) -> str:
        """Windows AHK driver"""
        out = load_tpl(self, ".ahk")
        out = substitute_lines(out, "LAYOUT", ahk_keymap(self))
        out = substitute_lines(out, "ALTGR", ahk_keymap(self, True))
        out = substitute_lines(out, "SHORTCUTS", ahk_shortcuts(self))
        return out

    @property
    def klc(self) -> str:
        """Windows driver (warning: requires CR/LF + UTF16LE encoding)"""
        if len(self.meta["name8"]) > 8:
            raise ValueError("`name8` max length is 8 charaters")

        # check `version` format
        # it must be `a.b.c[.d]`
        version = re.compile(r"^\d+\.\d+\.\d+(\.\d+)?$")
        if version.match(self.meta["version"]) is None:
            raise ValueError("`version` must be in `a.b.c[.d]` form")
        locale = self.meta["locale"]
        langid = get_langid(locale)
        out = load_tpl(self, ".klc")
        out = substitute_lines(out, "LAYOUT", klc_keymap(self))
        out = substitute_lines(out, "DEAD_KEYS", klc_deadkeys(self))
        out = substitute_lines(out, "DEAD_KEY_INDEX", klc_dk_index(self))
        out = substitute_token(out, "localeid", f"0000{langid}")
        out = substitute_token(out, "locale", locale)
        out = substitute_token(out, "encoding", "utf-16le")
        return out

    @property
    def klc_rc(self) -> str:
        """Windows resource file for C drivers"""
        out = load_tpl(self, ".RC")
        # version numbers are in "a,b,c,d" format
        version = self.meta["version"].replace(".", ",")
        out = substitute_token(out, "rc_version", version)
        return out

    @property
    def klc_c(self) -> str:
        """Windows keymap file for C drivers"""
        out = load_tpl(self, ".C")
        out = substitute_lines(out, "LAYOUT", c_keymap(self))
        out = substitute_lines(out, "DEAD_KEYS", c_deadkeys(self))
        out = substitute_lines(out, "DEAD_KEY_INDEX", c_dk_index(self))
        return out

    @property
    def xkb_keymap(self) -> str:  # will not work with Wayland
        """GNU/Linux driver (standalone / user-space)"""
        out = load_tpl(self, ".xkb_keymap")
        out = substitute_lines(out, "LAYOUT", xkb_keymap(self, xkbcomp=True))
        return out

    @property
    def xkb_symbols(self) -> str:
        """GNU/Linux driver (xkb patch, system or user-space)"""
        out = load_tpl(self, ".xkb_symbols")
        out = substitute_lines(out, "LAYOUT", xkb_keymap(self, xkbcomp=False))
        return out

    ###
    # JSON output: keymap (base+altgr layers) and dead keys
    #

    @property
    def json(self) -> Dict:
        """JSON layout descriptor"""
        return {
            "name": self.meta["name"],
            "description": self.meta["description"],
            "geometry": self.meta["geometry"].lower(),
            "keymap": web_keymap(self),
            "deadkeys": web_deadkeys(self),
            "altgr": self.has_altgr,
        }

    ###
    # SVG output
    #

    @property
    def svg(self) -> ET.ElementTree:
        """SVG drawing"""

        svg_ns = "http://www.w3.org/2000/svg"
        ET.register_namespace("", svg_ns)
        ns = {"": svg_ns}

        # Parse the SVG template
        res = pkgutil.get_data(__package__, "tpl/x-keyboard.svg")
        if res is None:
            return ET.ElementTree()
        svg = ET.ElementTree(ET.fromstring(res.decode("utf-8")))

        def set_key_label(key: ET.Element, lvl: int, char: str) -> None:
            for label in key.findall(f'g/text[@class="level{lvl}"]', ns):
                if char not in deadkeys:
                    label.text = char
                else:  # only show last char for deadkeys
                    if char == ODK_ID:
                        label.text = "★"
                    elif char == "*µ":
                        label.text = "α"
                    else:
                        label.text = char[-1]
                    label.set("class", f"{label.get('class')} deadKey")

        # Get layout data
        keymap = web_keymap(self)
        deadkeys = web_deadkeys(self)

        # Fill-in with layout
        for name, chars in keymap.items():
            for key in svg.findall(f'.//g[@id="{name}"]', ns):

                # Print 1-4 level chars
                for level_num, char in enumerate(chars, start=1):
                    if level_num == 1:
                        if chars[1] == upper_key(chars[0], blank_if_obvious=False):
                            continue
                    if level_num == 4:
                        if chars[3] == upper_key(chars[2], blank_if_obvious=False):
                            continue
                    set_key_label(key, level_num, char)

                # Print 5-6 levels (1dk deadkeys)
                if deadkeys and (odk := deadkeys.get(ODK_ID)):
                    level5 = odk.get(chars[0])
                    level6 = odk.get(chars[1])
                    if level5:
                        set_key_label(key, 5, level5)
                    if level6 and level6 != upper_key(level5, blank_if_obvious=False):
                        set_key_label(key, 6, level6)

        return svg



from pathlib import Path

from lxml import etree


def check_keylayout(filename: str):
    path = Path(__file__).parent.parent / f"dist/{filename}.keylayout"
    tree = etree.parse(path, etree.XMLParser(recover=True))
    dead_keys = []

    # check all keymaps/layers: base, shift, caps, option, option+shift
    for keymap_index in range(5):
        keymap_query = f'//keyMap[@index="{keymap_index}"]'
        keymap = tree.xpath(keymap_query)
        assert len(keymap) == 1, f"{keymap_query} should be unique"

        # check all key codes for this keymap / layer
        # (the key codes below are not used, I don't know why)
        excluded_keys = [
            54,
            55,
            56,
            57,
            58,
            59,
            60,
            61,
            62,
            63,
            68,
            73,
            74,
            90,
            93,
            94,
            95,
        ]
        for key_index in range(126):
            if key_index in excluded_keys:
                continue

            # ensure the key is defined and unique
            key_query = f'{keymap_query}/key[@code="{key_index}"]'
            key = tree.xpath(key_query)
            assert len(key) == 1, f"{key_query} should be unique"

            # ensure the key has either a direct output or a valid action
            action_id = key[0].get("action")
            if action_id:
                if action_id.startswith("dead_"):
                    dead_keys.append(action_id[5:])
                action_query = f'//actions/action[@id="{action_id}"]'
                action = tree.xpath(action_query)
                assert len(action) == 1, f"{action_query} should be unique"
                assert (
                    len(action_id) > 1
                ), f"{key_query} should have a multi-char action ID"
            else:
                assert (
                    len(key[0].get("output")) <= 1
                ), f"{key_query} should have a one-char output"

    # check all dead keys
    # TODO: ensure there are no unused actions or terminators
    for dk in dead_keys:

        # ensure all 'when' definitions are defined and unique
        when_query = f'//actions/action[@id="dead_{dk}"]/when'
        when = tree.xpath(when_query)
        assert len(when) == 1, f"{when_query} should be unique"
        assert when[0].get("state") == "none"
        assert when[0].get("next") == dk

        # ensure all terminators are defined and unique
        terminator_query = f'//terminators/when[@state="{dk}"]'
        terminator = tree.xpath(terminator_query)
        assert len(terminator) == 1, f"{terminator_query} should be unique"
        assert len(terminator[0].get("output")) == 1


def test_keylayouts():
    check_keylayout("q-ansi")
    check_keylayout("q-intl")
    check_keylayout("q-prog")



from pathlib import Path
from typing import Dict, List

from .layout import KeyboardLayout
from .template import SCAN_CODES
from .utils import Layer, load_data

SEPARATOR = (
    "--------------------------------------------------------------------------------"
)

MARKDOWN_HEADER = """Defining a Keyboard Layout
================================================================================

Kalamine keyboard layouts are defined with TOML files including this kind of
ASCII-art layer templates:

```KALAMINE_LAYOUT```
"""

TOML_HEADER = """# kalamine keyboard layout descriptor
name        = "qwerty-custom"  # full layout name, displayed in the keyboard settings
name8       = "custom"         # short Windows filename: no spaces, no special chars
locale      = "us"             # locale/language id
variant     = "custom"         # layout variant id
author      = "nobody"         # author name
description = "custom QWERTY layout"
url         = "https://OneDeadKey.github.com/kalamine"
version     = "0.0.1"
geometry    = """

TOML_FOOTER = """
[spacebar]
1dk         = "'"  # apostrophe
1dk_shift   = "'"  # apostrophe"""


def dead_key_table() -> str:
    out = f"\n    id  XKB name          base -> accented chars\n    {SEPARATOR[4:]}"
    for item in load_data("dead_keys"):
        if (item["char"]) != "**":
            out += f"\n    {item['char']}  {item['name']:<17} {item['base']}"
            out += f"\n                       -> {item['alt']}"
    return out


def core_guide() -> List[str]:
    sections: List[str] = []

    for title, content in load_data("user_guide").items():
        out = f"\n{title.replace('_', ' ')}\n{SEPARATOR}"

        if isinstance(content, dict):
            for subtitle, subcontent in content.items():
                out += f"\n\n### {subtitle.replace('_', ' ')}"
                out += f"\n\n{subcontent}"
                if subtitle == "Standard_Dead_Keys":
                    out += dead_key_table()
        else:
            out += f"\n\n{content}"

        sections.append(out)

    return sections


def dummy_layout(
    geometry: str = "ISO",
    altgr: bool = False,
    odk: bool = False,
    meta: Dict[str, str] = {},
) -> KeyboardLayout:
    """Create a dummy (QWERTY) layout with the given characteristics."""

    # load the descriptor, but only keep the layers we need
    descriptor = load_data("layout")
    if not altgr:
        del descriptor["altgr"]
    if not odk:
        del descriptor["1dk"]
        descriptor["base"] = descriptor.pop("alpha")
    else:
        del descriptor["alpha"]
        descriptor["base"] = descriptor.pop("1dk")

    # XXX this should be a dataclass
    for key, val in meta.items():
        descriptor[key] = val

    # make a KeyboardLayout matching the input parameters
    descriptor["geometry"] = "ANSI"  # layout.yaml has an ANSI geometry
    layout = KeyboardLayout(descriptor)
    layout.geometry = geometry

    # ensure there is no empty keys (XXX maybe this should be in layout.py)
    for key in SCAN_CODES["web"].keys():
        if key not in layout.layers[Layer.BASE].keys():
            layout.layers[Layer.BASE][key] = "\\"
            layout.layers[Layer.SHIFT][key] = "|"

    return layout


def draw_layout(geometry: str = "ISO", altgr: bool = False, odk: bool = False) -> str:
    """Draw a ASCII art description of a default layout."""

    # make a KeyboardLayout, just to get the ASCII arts
    layout = dummy_layout(geometry, altgr, odk)

    def keymap(layer_name: str) -> str:
        layer = "\n".join(getattr(layout, layer_name))
        return f"\n{layer_name} = '''\n{layer}\n'''\n"

    content = ""
    if odk:
        content += keymap("base")
        if altgr:
            content += keymap("altgr")
    elif altgr:
        content += keymap("full")
    else:
        content += keymap("base")

    return content


###
# Public API
##


def user_guide() -> str:
    """Create a user guide with a sample layout description."""

    header = MARKDOWN_HEADER.replace(
        "KALAMINE_LAYOUT", draw_layout(geometry="ANSI", altgr=True)
    )
    return header + "\n" + "\n".join(core_guide())


def create_layout(output_file: Path, geometry: str, altgr: bool, odk: bool) -> None:
    """Create a new TOML layout description."""

    content = f'{TOML_HEADER}"{geometry.upper()}"\n'
    content += draw_layout(geometry, altgr, odk)
    if odk:
        content += TOML_FOOTER

    for topic in core_guide():
        content += f"\n\n\n# {SEPARATOR}"
        content += "\n# ".join(topic.rstrip().split("\n"))

    with open(output_file, "w", encoding="utf-8", newline="\n") as file:
        file.write(content.replace(" \n", "\n"))
